"""
Processing Class, holds list of data processing steps
"""

# Standard library modules
import warnings
import logging
import json

# Non-standard modules

# obsinfo modules

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Processing(object):
    """
    No equivalent class in obspy/StationXML

    Saves a list of Processing steps as strings
    For now, just stores the list. Will be converted to StationXML comments

    Attributes:
        processing_list (list): list of processing steps with attributes,
            either linear_drift or leapsecond

    """

    def __init__(self, attributes):
        """
        Constructor

        Args:
        attributes (list): list of processing steps (linear_drift or
            leapsecond) with attributes
        """

        self.processing_list = []

        if not attributes:
            return

        # make it a list for standard processing if user forgot the dash
        if not isinstance(attributes, list):
            attributes = [attributes]

        # Wayne added self.attributes to allow makescript_LC2SDS
        self.attributes = attributes

        for attr in attributes:
            self.processing_list.append(json.dumps(attr))

    def __repr__(self):
        s = f'Processing({self.processing_list})'
        return s



"""
Instrument and Operator classes
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.response import (Response, InstrumentSensitivity
                                           as obspy_Sensitivity)

# obsinfo modules
from .instrument_component import (InstrumentComponent)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Instrument(object):    # Was a Channel subclass, but don't see why
    """
    An instrument is an ensemble of a sensor, a datalogger and possibly a
    preamplifier. It also includes a selected configuration for each one of
    these instrument components.

    Attributes:
        datalogger (:class:`Datalogger`)
        sensor: (:class:`Sensor`)
        preamplifier: (:class:`Preamplifier`)
        sample_rate (float): from datalogger sample rate
        delay_correction (float): from datalogger delay correction
        seed_band_base_code (str): from sensor band base code
        seed_instrument_code (str): from sensor instrument code
        seed_orientation (str): from orientation in the channel itself
    """

    def __init__(self, attributes_dict, channel_modif={}):
        """
        Constructor

        :param attributes_dict_or_list: instrument attributes
        :type attributes_dict_or_list: dictionary/ObsMetadata object or list
            of dictionaries//ObsMetadata objects
        :param channel_modif: modification of attributes per channel specified
            in stations
        :type channel_modif: dict or object of :class:`ObsMetadata`
        """
        self.correction = None
        self.delay = None

        if not attributes_dict:
            msg = 'No instrument attributes'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError()

        datalogger_config_selector = attributes_dict.get_configured_element(
            'datalogger_configuration', channel_modif, {}, None)
        sensor_config_selector = attributes_dict.get_configured_element(
            'sensor_configuration', channel_modif, {}, None)
        preamplifier_config_selector = attributes_dict.get_configured_element(
            'preamplifier_configuration', channel_modif, {}, None)

        self.sensor = InstrumentComponent.dynamic_class_constructor(
            'sensor',
            attributes_dict, channel_modif, sensor_config_selector)
        self.preamplifier = InstrumentComponent.dynamic_class_constructor(
            'preamplifier',
            attributes_dict, channel_modif, preamplifier_config_selector)
        self.datalogger = InstrumentComponent.dynamic_class_constructor(
            'datalogger',
            attributes_dict, channel_modif, datalogger_config_selector)
        # add the three component response stages
        self.combine_response_stages()

        # Validate inputs and outputs of complete response sequence and
        # correct delay
        self.integrate_response_stages()
        self.obspy_stages = [x.to_obspy() for x in self.response_stages]
        self.obspy_response = self.to_obspy()
        self.add_sensitivity(self.obspy_response)

    def __repr__(self):
        return f'\nInstrument(Polarity="{self.polarity}", '\
               f'Output sample rate={self.total_output_sample_rate})'

    def combine_response_stages(self):
        """
        Adds all response stages as obsinfo and obpsy objects and renumbers
        them
        """
        response_st = self.sensor.response_stages.stages

        if self.preamplifier and self.preamplifier.response_stages:
            response_st += self.preamplifier.response_stages.stages

        response_st += self.datalogger.response_stages.stages

        # Order the stage_sequence_numbers
        for i in range(len(response_st)):
            response_st[i].stage_sequence_number = i+1

        self.response_stages = response_st

    def integrate_response_stages(self):
        """
        Integrates the stages with one another

        1) Renumber stages sequentially
        2) Verify/set units and sample rates
        3) Assure same frequency is used for consecutive PZ filters
        4) Calculate global polarity of the whole set of response stages
        5) Set global response delay correction
        6) Validate sample_rate expressed in datalogger component is equal to
           global response sample rate
        """
        # Stack response stages
        stages = self.response_stages

        prev_stage = stages[0]
        prev_frequency = prev_stage.filter.normalization_frequency \
            if prev_stage.filter.type == 'PolesZeros' else None
        prev_polarity = prev_stage.polarity
        accum_sample_rate = prev_stage.output_sample_rate

        for this_stage in stages[1:]:
            # 2a) Verify continuity of units
            if prev_stage.output_units != this_stage.input_units:
                msg = f'Stage {prev_stage.stage_sequence_number} and '\
                      f'{this_stage.stage_sequence_number} units don\'t match'
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)

            # 2b) Verify/set continuity of sample rate
            if prev_stage.input_sample_rate:
                if not this_stage.decimation_factor:
                    msg = f'No decimation factor for stage '\
                          f'{this_stage.stage_sequence_number}, assuming = 1'
                    warnings.warn(msg)
                    logger.warning(msg)
                    this_stage.decimation_factor = 1
                next_input_rate = prev_stage.input_sample_rate \
                    / this_stage.decimation_factor

                if this_stage.input_sample_rate:
                    if this_stage.output_sample_rate != next_input_rate:
                        msg = f'stage {this_stage.stage_sequence_number} '\
                              f'sample rate {this_stage.output_sample_rate} '\
                              f'!= expected {next_input_rate}'
                        warnings.warn(msg)
                        logger.error(msg)
                        raise ValueError(msg)
                else:
                    this_stage.input_sample_rate = accum_sample_rate

            # 2c) Calculate/verify delay and correction
            if this_stage.input_sample_rate:
                this_stage.calculate_delay()
            if self.correction is None and self.delay is not None:
                self.correction = self.delay

            # 3) Station XML requires that all PZ stages have the same
            #    normalization frequency.  Check this condition
            if prev_frequency and prev_frequency != 0 \
                    and this_stage.filter.type == 'PolesZeros':
                if prev_frequency != this_stage.filter.normalization_frequency:
                    msg = 'Normalization frequencies for PZ stages '\
                          f'{prev_stage.stage_sequence_number} and '\
                          f'{this_stage.stage_sequence_number} don\'t match'
                    warnings.warn(msg)
                    logger.warning(msg)
                prev_frequency = this_stage.filter.normalization_frequency

            # 4) Calculate global polarity
            if not this_stage.polarity:  # default polarity is positive
                this_stage.polarity = 1

            prev_polarity *= this_stage.polarity

            prev_stage = this_stage
            prev_frequency = prev_stage.filter.normalization_frequency \
                if prev_stage.filter.type == 'PolesZeros' else None
            accum_sample_rate = prev_stage.output_sample_rate

        # 5) Out of the loop: check global output sample rate
        total_expected_sample_rate = accum_sample_rate

        if not total_expected_sample_rate == self.sample_rate:
            msg = f'Datalogger declared sample rate {self.sample_rate} is '\
                  'different from calculated overall sample rate of stages '\
                  f'{total_expected_sample_rate}'
            warnings.warn(msg)
            logger.warning(msg)

        # Set global response attributes
        self.polarity = prev_polarity
        self.total_output_sample_rate = accum_sample_rate

    def to_obspy(self):
        """
        Return equivalent obspy class

        Returns:
            ():class:`obspy.core.inventory.response.Response`)
        """
        return Response(resource_id=None,
                        instrument_sensitivity=None,
                        instrument_polynomial=None,
                        response_stages=self.obspy_stages)

    def add_sensitivity(self, obspy_response):
        """
        Adds sensitivity to an obspy Response object
        Based on ..misc.obspy_routines.response_with_sensitivity

        Args:
            obspy_response (:class:`obspy.core.inventory.response.Response`):
        """

        response_stg = self.response_stages
        true_input_units = input_units = response_stg[0].input_units
        output_units = response_stg[-1].output_units

        if "PA" in true_input_units.upper():
            # MAKE OBSPY THINK ITS M/S TO CORRECTLY CALCULATE SENSITIVITY
            input_units = "M/S"
        gain_prod = 1.
        for stage in response_stg:
            gain_prod *= stage.gain

        sensitivity = obspy_Sensitivity(
            gain_prod,
            # gain_frequency could be provided, according to StationXML, but we
            # assume it's equal to the gain frequency of first stage
            response_stg[0].gain_frequency,
            input_units=input_units,
            output_units=output_units,
            input_units_description=response_stg[0].input_units_description,
            output_units_description=response_stg[-1].output_units_description
            )

        obspy_response.instrument_sensitivity = sensitivity
        # obspy_response.recalculate_overall_sensitivity(sensitivity.frequency)
        obspy_response.instrument_sensitivity.input_units = true_input_units
        obspy_response.instrument_sensitivity.output_units = output_units

    def get_response_stage(self, num):
        """
        Returns the response stage in a given position

        Args:
            num (int): stage number, starting with zero and ordered from
                sensor to datalogger
        """
        # All response stages are at the instrument_component level
        stages = self.response_stages
        assert(num <= stages[-1].stage_sequence_number), \
            'response stage out of range: {num}'
        return stages[num]

    @property
    def equipment_datalogger(self):
        return self.datalogger.equipment

    @property
    def equipment_sensor(self):
        return self.sensor.equipment

    @property
    def equipment_preamplifier(self):
        return self.preamplifier.equipment

    @property
    def sample_rate(self):
        return self.datalogger.sample_rate

    @property
    def delay_correction(self):
        return self.delay_correction

    @property
    def seed_band_base_code(self):
        return self.sensor.seed_band_base_code

    @property
    def seed_instrument_code(self):
        return self.sensor.seed_instrument_code

    @property
    def seed_orientation(self):
        """
        Same as orientation. Kept for compatibility.
        """
        return self.orientation



import gitlab
import urllib
import json
import base64
import re
import os
from pathlib import Path
from urllib import parse as urlparse
from urllib.parse import unquote
from urllib.request import urlopen
import logging
logger = logging.getLogger("obsinfo")

# obsinfo
from .configuration import ObsinfoConfiguration

# private token or personal token authentication


class gitLabFile(object):
    """
    Provide the methods to use the gitlab API to read a remote file and decode it
    
      **Attributes:**
     
       
       *None*
       
    """
      
    @staticmethod
    def get_gitlab_file(uri, read_file=True):
        """
        Get the file content pointed at by uri and decode it.
        
        Uses b64 first to get the remote file and convert to a byte string, 
        and then utf-8 to convert to regular string.
        
        :param uri: uri to read
        :type uri: string or path-like object
        :param read_file: If true, reads the content. If not, simply checks the file exists
        :returns: string of read content
        :raises: FileNotFoundError, ValueError
        
        """
        
        # Use configuration ifnormation
        gitlab_repository = ObsinfoConfiguration.Instance().gitlab_repository
        project_path = ObsinfoConfiguration.Instance().gitlab_project
        obsinfo_branch = ObsinfoConfiguration.Instance().obsinfo_branch
        
        if not gitlab_repository or not project_path:
            msg = "One or several environment variables, OBSINFO_GITLAB_REPOSITORY or \
                     OBSINFO_PROJECTPATH in the configuration file ~/.obsinforc, are missing"
            logger.error(msg)
            raise ValueError(msg)
        
        with gitlab.Gitlab(gitlab_repository) as gl:
        
            project = gl.projects.get(project_path)
            if not project:
                msg = f'Project not found in repository: {project_path}'
                logger.error(msg)
                raise FileNotFoundError(msg)    
            
            if urlparse.urlsplit(uri).scheme: #Sometime it comes with scheme, sometimes not
                uri = urlparse.urlsplit(uri).path
                pattern = "/"
            else:
                # Remove elements from path
                pat1 = re.compile("http://")
                pat2 = re.compile("https://")
                if pat1.match(gitlab_repository):
                    gitlab_repository = pat1.sub("", gitlab_repository)
                if pat2.match(gitlab_repository):
                    gitlab_repository = pat2.sub("", gitlab_repository)
                
                pattern = ("/" if gitlab_repository[0] != "/" else "") + gitlab_repository 
            
                if gitlab_repository[-1] != "/":
                    pattern += "/"
            
            #we assume project_path has no leading slash
            pattern += project_path
            if project_path[-1] != "/":
                pattern += "/"
                          
            uri = re.sub(pattern, "", uri) # remove everything but path relative to project path
            
            try:
                f = project.files.get(file_path=uri, ref=obsinfo_branch)
                if not read_file:
                    return True
            except gitlab.exceptions.GitlabOperationError:
                if read_file:
                    msg = f'Error reading remote (gitlab) file: {uri}'
                    print(msg)
                    raise FileNotFoundError(msg)               
                else:
                    return None
            # get the decoded content. Two decodes are necessary, from 8character string to bytes and 
            # from bytes to a python string using utf-8
            bytecontent = base64.b64decode(f.content)
            ret = bytecontent.decode("utf-8")

        return ret
    
    @staticmethod
    def isRemote(file):
        """
        Checks if scheme means file is remote.
        
        :param file: filename to be checked, with complete uri
        :type file: str
        :returns: boolean. True if remote, False otherwise
        """
        return re.match('^http:', file) or re.match('^https:', file)





"""
Network :class:

"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.network import Network as obspy_Network
from obspy.core.inventory.util import (Person, Comment, PhoneNumber)
from obspy.core.utcdatetime import UTCDateTime

# obsinfo modules
from .station import (Station)
from .operator_class import Operator

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Network(object):
    """
    Network obsinfo: Equivalent to obspy/StationXML Network

    Attributes:
         campaign_ref (str)
         fdsn_code (str)
         fdsn_name (str)
         start_date (str with date format)
         end_date (str with date format)
         description  (str)
         restricted_status (str)
         operator (object of :class:`.Operator`)
         stations (list of objects of :class:`.Station`)
         comments (list of str)
         extras (list of str)
         obspy_network (:class~`obspy.core.inventory.network.Network):
           obspy equivalent of this class
    """

    def __init__(self, attributes_dict=None, station_only=False):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): dictionary from
                network info file
            station_only (bool): Instructs class Station to create object
                with no instrumentation
        Raises:
            (TypeError): if attributes_dict is empty
        """
        if not attributes_dict:
            msg = 'No network attributes'
            logger.error(msg)
            raise TypeError(msg)

        self.campaign_ref = attributes_dict.get("campaign_ref_name", None)
        network_info = attributes_dict.get("network_info", None)

        if network_info:
            self.fdsn_code = network_info.get("code", None)
            self.fdsn_name = network_info.get("name", None)
            self.start_date = network_info.get("start_date", 0)
            self.end_date = network_info.get("end_date", 0)
            self.description = network_info.get("description", None)

        self.restricted_status = attributes_dict.get("restricted_status", None)
        self.operator = Operator(attributes_dict.get("operator", None))
        st_op = attributes_dict.get("stations_operator", None)
        if st_op is None:
            stations_operator = self.operator
        else:
            stations_operator = Operator(st_op)

        stations = attributes_dict.get("stations", None)
        self.stations = [Station(k, v, station_only, stations_operator)
                         for k, v in stations.items()]

        self.comments = attributes_dict.get("comments", [])
        self.extras = [str(k) + ": " + str(v)
                       for k, v in (attributes_dict.get('extras', {})).items()]
        self.convert_notes_and_extras_to_obspy()

        self.obspy_network = self.to_obspy()

    def __repr__(self):
        s = f'Network(Campaign: {self.campaign_ref}, '
        s += f'FDSN Code: {self.fdsn_code}, FDSN Name: {self.fdsn_code}, '
        s += f'Start date: {self.start_date}, End date: {self.end_date}, '
        s += 'Description: {self.description}, '
        s += f'{len(self.stations)} stations)'
        return s

    def to_obspy(self):
        """
         Convert network object to obspy object

         Returns:
            (:class:~obspy.core.inventory.network.Network): corresponding
                obspy Network
        """
        # This should be done in the Operator class, not here
        if self.operator:
            operator = self.operator.to_obspy()
        else:
            operator = None

        stations_number = len(self.stations)
        start_date = UTCDateTime(self.start_date) if self.start_date else None
        end_date = UTCDateTime(self.end_date) if self.end_date else None
        comments = [Comment(s) for s in self.comments]

        self.obspy_network = obspy_Network(
            code=self.fdsn_code,
            stations=[st.obspy_station for st in self.stations],
            total_number_of_stations=stations_number,
            selected_number_of_stations=stations_number,
            description=self.fdsn_name + " - " + self.description,
            comments=comments,
            start_date=start_date,
            end_date=end_date,
            restricted_status=self.restricted_status,
            alternate_code=None,
            historical_code=None,
            data_availability=None,
            identifiers=None,
            operators=[operator],
            source_id=None)
#         for st in self.stations:  # complete stations
#             st.operators = [operator]
        return self.obspy_network

    def convert_notes_and_extras_to_obspy(self):
        """
        Convert info file notes and extras to XML comments
        """
        if self.extras:
            self.comments.append('EXTRA ATTRIBUTES (for documentation only):')
            if isinstance(self.extras, list):
                self.comments.extend(self.extras)
            else:
                self.comments.append(self.extras)



"""
 Main functions for obsinfo-makeStationXML
 
 Creates obsinfo objects starting with a network object in a hierarchy which 
 strongly follows the hierarchy of StationXML files.
 Then converts the objects to a StationXML file using obspy.

"""

# General library imports
import sys
import os
import re
import warnings

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
from pathlib import Path, PurePath
from json.decoder import JSONDecodeError
import logging
from logging.handlers import RotatingFileHandler

from argparse import ArgumentParser

#Third party imports
import obspy
from obspy.core.inventory import Inventory,  Station, Channel, Site
from obspy.core.inventory import Network as obspy_Network
from obspy.clients.nrl import NRL
#obsinfo imports
from ..network import (Network)
from ..obsMetadata.obsmetadata import (ObsMetadata)
from ..misc.discoveryfiles import (Datapath)
from ..print_version import main as print_version 
import obsinfo
from ..misc.const import *
from ..misc.configuration import ObsinfoConfiguration

def main(argv=None, dp=None):
    """
    Entry point for obsinfo-makeStationXML. 
     
     1) Setups status variables from command line arguments.
     2) Read yaml or jason file invoking read_info_file, which returns a dictionary. Optionally validates dictionary schema.
     3) Creates obsinfo objects starting from network object from the dictionary 
     4) Converts these to StationXML using obpsy libraries. 
     
    Manages all uncaught exceptions.
    
    Args:
        argv (list): list of command-line arguments to pass to ArgumentParser.
            If None, will use sys.argv
        dp (Datapath): Datapath object specifying where to look for files.
            If None, will use values specified in .obsinforc
    """
    
    # create list of directories to search for files    
    if dp == None:
        dp = Datapath()
    args = retrieve_arguments(argv, dp)
    logger = init_logging()
    
    if args.verbose:
        print(f'Using OBSINFO_DATAPATH: {dp.datapath_list}')
    
    logger.info(f'Using OBSINFO_DATAPATH: {dp.datapath_list}') 
   
    _make_StationXML(logger, args, dp)

    if argv is None:
        sys.exit(EXIT_SUCCESS)
    
  
def _make_StationXML(logger, args, dp):
    # try:               
                     
    file = Path(args.input_filename).name    
   
    # if args.validate:
    #     if args.verbose:
    #         print(f'Validating network file: {file}')
    #     logger.info(f'Validating network file: {file}')    
    #         
    #     ret = ObsMetadata().validate(args.schemapath,  str(args.input_filename), 
    #                                  remote=args.remote, 
    #                                  format="yaml", 
    #                                  type="network", 
    #                                  verbose=args.verbose, 
    #                                  schema_file="network", 
    #                                  quiet=False)
                
    info_dict = ObsMetadata.read_info_file(args.input_filename, dp, 
                                           remote=args.remote,
                                           verbose=args.verbose)
    
    net_dict = info_dict.get('network', None)
    if not net_dict:
        return 
            
    if args.verbose:
        print(f'Processing network file: {file}')
    logger.info(f'Processing network file: {file}')
    
    obj = Network(ObsMetadata(net_dict), args.station)
              
    if args.verbose:
       print(f'Network file parsed successfully for: {file}')
    logger.info(f'Network file parsed successfully for: {file}')
       
    networks=[obj.obspy_network]
    if not isinstance(obj.obspy_network, obspy_Network):
        print("Not a network object")
        logger.error("Not a network object")
        
    if not args.quiet:
       print(obj.obspy_network)
    
    logger.info(obj.obspy_network)
        
    version = os.environ.get('OBSINFO_VERSION')
    
    author = get_first_author(info_dict)

    inv = Inventory(networks, # module=version,
                    module_uri="https://gitlab.com/resif/obsinfo",
                    sender=author, source="ObsPy")

    if not args.test: # Generate Stationxml file
        if not args.output:
            stem_name = Path(file).stem      # remove .yaml
            stem_name = Path(stem_name).stem # Remove .network
            output_filename = stem_name + ".station.xml"
        else:
            output_filename = args.output
        
        stationxml=inv.write(output_filename, format="stationxml", validate=False)
     
    if not args.quiet and not args.test:
       print(f'StationXML file created successfully: {output_filename}') 
    logger.info(f'StationXML file created successfully: {output_filename}')          
               
    # except TypeError:
    #     print("Illegal format: fields may be missing or with wrong format in input file, or there is a programming error")
    #     logger.error("TypeError: Illegal format: fields may be missing or with wrong format in input file, or there is a programming error")
    #     if args.debug:
    #         raise
    #     
    #     sys.exit(EXIT_DATAERR)
#     except (KeyError, IndexError):
#         print("Illegal value in dictionary key or list index")
#         logger.error("KeyError, IndexError: Illegal value in dictionary key or list index")
#         if args.debug:
#             raise
#         sys.exit(EXIT_SOFTWARE)
#     except ValueError:
#         print("An illegal value was detected")
#         logger.error("ValueError: An illegal value was detected")
#         if args.debug:
#             raise
#         sys.exit(EXIT_DATAERR)
#     except FileNotFoundError:
#         if args.debug:
#             raise
#         print("File could not be found")
#         logger.error("FileNotFoundError: File could not be found")
#         sys.exit(EXIT_NOINPUT)
#     except JSONDecodeError:
#         print("File and/or subfiles have an illegal format. Probably indentation or missing quotes/parentheses/brackets")
#         logger.error("JSONDecodeError: File and/or subfiles have an illegal format. Probably indentation or missing quotes/parentheses/brackets")
#         if args.debug:
#             raise 
#         sys.exit(EXIT_DATAERR)
#     except (IOError, OSError, LookupError):
#         print("File could not be opened or read")
#         logger.error("IOError, OSError, LookupError: File could not be opened or read")
#         if args.debug:
#             raise
#         sys.exit(EXIT_UNAVAILABLE)
#     except AttributeError:
#         print("Programming error: an object in code had a wrong attribute")
#         logger.debug("AttributeError: Programming error: an object in code had a wrong attribute")
#         if args.debug:
#             raise
#         sys.exit(EXIT_SOFTWARE)
#     except:
#         print("General exception")
#         logger.debug("General exception")
#         if args.debug:
#             raise
#         sys.exit(EXIT_FAILURE)


def retrieve_arguments(argv, datapath):
    
    """
    Retrieve arguments from command line. Setup several status variables and get information file name
      
    Args:
        argv (list): command line arguments.  If None, uses sys.argv
        datapath (:class:`.Datapath`): Object containing paths to find
            repository files, read from .obsinforc
    Returns:
        args (NameSpace): All status variables and the information file name.
    """
            
    # Parse the arguments
    
    parser_args = ArgumentParser(prog="obsinfo-makeStationXML")
    
    #flags
    parser_args.add_argument(
        "-r", "--remote", action='store_true', default=False,
        help="Assumes input filename is discovered through OBSINFO_DATAPATH "
             "environment variable. Does not affect treatment of $ref in info files")
    # parser_args.add_argument("-l", "--validate", action='store_true', default=None,
    #                          help="Performs complete validation, equivalent to obsinfo-validate, before processing")
    parser_args.add_argument("-v", "--verbose", action='store_true', default=False,
                             help="Prints processing progression")
    parser_args.add_argument("-q", "--quiet", action='store_true', default=False,                    
                             help="Silences a human-readable summary of processed information file")
    parser_args.add_argument("-d", "--debug",  action='store_true', default=False,
                             help="Turns on exception traceback") 
    parser_args.add_argument("-t", "--test",  action='store_true', default=False,
                             help="Produces no output") 
    parser_args.add_argument("-V", "--version",  action="store_true", default=False,
                             help="Print the version and exit") 
    parser_args.add_argument("-S", "--station",  action="store_true", default=False,
                             help="Create a StationXML file with no instrumentation") 
    # optional arguments
    parser_args.add_argument("-o", "--output", default=None,
                             help="Names the output file. Default is <input stem>.station.xml")
    # positional arguments
    parser_args.add_argument("input_filename", type=str,  nargs=1, 
                             help="is required and must be a single value") 
    
    if argv is not None:
        args = parser_args.parse_args(argv) 
    else:
        args = parser_args.parse_args() 
    
    if args.version:
        print_version() 
        sys.exit(EXIT_SUCCESS) 

    # schemas must always be installed under obsinfo/data/schemas   
    args.schemapath = Path(obsinfo.__file__).parent.joinpath('data', 'schemas')
    
    if not args.input_filename:
        print("No input filename specified")
        sys.exit(EXIT_USAGE) 
        
    input_filename = args.input_filename[0]
    
    args.input_filename = str(datapath.build_datapath(input_filename) if args.remote \
                         else Path(os.getcwd()).joinpath(input_filename))
    
    return args

def init_logging():
    """
    Create or open a rotating logging file and add it to ObsinfoConfiguration
    
    :returns: object of Logger class 
    """
    
    logfile = Path.home().joinpath('.obsinfo', 'obsinfolog-makeStationXML')
    
    
    logger = logging.getLogger("obsinfo")
    
    logger.setLevel(logging.DEBUG)
    # add a rotating handler with 200K (approx) files and just two files
    handler = RotatingFileHandler(logfile, maxBytes=200000,
                                  backupCount=2)
    frmt = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(frmt)
    logger.addHandler(handler)  
    
    return logger 
        

def get_first_author(info_dict):
    """
    Get info file first author info and return it in a string.

    Used to fill the author attribute in the Inventory object of obspy.
    
    :param info_dict: The name to use.
    :type name: str.
    :param state: Current state to be in.
    :type state: bool.
    :returns:  str -- first author first name and last name and contact information
    """
    
    # Ojo, might change to parse arguments
    
    rev = info_dict.get('revision', None)
    if not rev:
        return ""
   
    authors = rev.get('authors', None)

    if authors:
        author = authors[0].get('first_name', "") + " " + authors[0].get('last_name', "") \
           + ", " + authors[0].get('institution', "") +  " Email: " + authors[0].get('email', "") + "  Phones: "
        for phn in authors[0].get('phones', ""):
            author += phn + " "
    
    return author 

    
if __name__ == '__main__':
    main()


"""
Filter class and subclasses
"""
# Standard library modules
import math as m
import warnings
import logging

# Non-standard modules
# import obspy.core.inventory.response as obspy_response
# from scipy._lib.doccer import extend_notes_in_docstring

# from ..misc.configuration import ObsinfoConfiguration
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Filter(object):
    """
    Filter is superclass of all filter classes

    Attributes:
        type (str): filter type
        offset (int): offset is samples to skip at the beggining of a signal
            in digital filters
    """

    def __init__(self, type="PolesZeros", offset=0):
        """
        Constructor

        Args:
            type (str): filter type
            offset (int): samples that an impulse is offset by a digital filter
        """
        self.type = type
        self.offset = offset

    @staticmethod
    def dynamic_class_constructor(attributes_dict, channel_modif={},
                                  selected_config={}, stage_id="-1",
                                  gain_frequency=1.):
        """
        Creates an appropriate Filter subclass from an attributes_dict

        Args:
            attributes_dict (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): the configuration
                description that will override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
            gain_frequency (float): frequency at which gain was specified.
                Used for PoleZeros Normalization factor/frequency
        Returns:
            (:class:`.Filter`): object of the adequate filter subclass
        Raises:
            (TypeError): if filter type is not valid
        """
        if attributes_dict is None:
            msg = "No attributes in filter"
            logger.error(msg)
            raise TypeError(msg)

        if "type" not in attributes_dict:
            msg = 'No "type" specified for filter in stage #{stage_id}'
            logger.error(msg)
            raise TypeError(msg)
        else:
            filter_type = attributes_dict.get_configured_element(
                'type', channel_modif, selected_config, None)
            if filter_type == 'PolesZeros':
                obj = PolesZeros.dynamic_class_constructor(
                    filter_type, attributes_dict, channel_modif,
                    selected_config, stage_id, gain_frequency)
            elif filter_type == 'FIR':
                obj = FIR.dynamic_class_constructor(
                    filter_type, attributes_dict, channel_modif,
                    selected_config, stage_id)
            elif filter_type == 'Coefficients':
                obj = Coefficients.dynamic_class_constructor(
                    filter_type, attributes_dict, channel_modif,
                    selected_config, stage_id)
            elif filter_type == 'ResponseList':
                obj = ResponseList.dynamic_class_constructor(
                    filter_type, attributes_dict, channel_modif,
                    selected_config, stage_id)
            elif filter_type == 'ADConversion':
                obj = AD_Conversion.dynamic_class_constructor(
                    filter_type, attributes_dict, channel_modif,
                    selected_config, stage_id)
            elif filter_type == 'Analog':
                obj = Analog.dynamic_class_constructor(
                    filter_type, attributes_dict, channel_modif,
                    selected_config, stage_id, gain_frequency)
            elif filter_type == 'Digital':
                obj = Digital.dynamic_class_constructor(
                    filter_type, attributes_dict, channel_modif,
                    selected_config, stage_id)
            else:
                msg = (f'Unknown Filter type: "{filter_type}" in '
                       'stage #{stage_id}')
                logger.error(msg)
                raise TypeError(msg)

        return obj


class PolesZeros(Filter):
    """
    PolesZeros filter

    Attributes:
        transfer_function_type (str): one of  'LAPLACE (RADIANS/SECOND)',
            'LAPLACE (HERTZ)','DIGITAL (Z-TRANSFORM)'
        poles (list of complex numbers)
        zeros (list of complex numbers)
        normalization_frequency (float)
        normalization_factor (float)
    """

    def __init__(self, filter_type,
                 transfer_function_type='LAPLACE (RADIANS/SECOND)',
                 poles=[], zeros=[], normalization_frequency=1.,
                 normalization_factor=None, offset=0, stage_id=-1):
        """
        Constructor

        Args:
            filter_type (str):
            transfer_function_type (str): one of 'LAPLACE (RADIANS/SECOND)',
                'LAPLACE (HERTZ)', 'DIGITAL (Z-TRANSFORM)'
            poles (list): complex numbers specified as [a, b]
            zeros (list): complex numbers specified as [a, b]
            normalization_frequency (float):
            normalization_factor (float):
            offset (int): makes no sens for an analog filter
            stage_id (int): id of corresponding stage. Used for reporting only
        """
        # poles and zeros should be lists of complex numbers
        if transfer_function_type not in ['LAPLACE (RADIANS/SECOND)',
                                          'LAPLACE (HERTZ)',
                                          'DIGITAL (Z-TRANSFORM)']:
            msg = (f'Illegal transfer_function_type in PolesZeros: '
                   f'"{transfer_function_type}" in stage #{stage_id}')
            logger.error(msg)
            raise TypeError(msg)

        self.transfer_function_type = transfer_function_type
        self.poles = poles
        self.zeros = zeros

        self.normalization_frequency = normalization_frequency
        if normalization_frequency and normalization_factor:
            self.normalization_factor = normalization_factor
        elif filter_type == 'Analog':
            self.normalization_factor = 1.0
        else:
            self.normalization_factor = self.calc_normalization_factor(stage_id)
        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict,
                                  channel_modif={},
                                  selected_config={}, stage_id=-1,
                                  gain_frequency=1.):
        """
        Create PolesZeros instance from an attributes_dict

        Args:
            filter_type (str): type of filter, necessarily 'PolesZeros'
            attribute_list (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): the configuration
                description that will override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
            gain_frequency (float): frequency at which gain was specified.
        Returns:
            (:class:`.PolesZeros`)
        """
        obj = cls(filter_type,
                  attributes_dict.get_configured_element(
                    'transfer_function_type', channel_modif, selected_config,
                    'LAPLACE (RADIANS/SECOND)'),
                  [(float(x[0]) + 1j*float(x[1]))
                   for x in attributes_dict.get_configured_element(
                        'poles', channel_modif, selected_config, [])],
                  [(float(x[0]) + 1j*float(x[1]))
                   for x in attributes_dict.get_configured_element(
                        'zeros', channel_modif,  selected_config, [])],
                  attributes_dict.get_configured_element(
                    'normalization_frequency', channel_modif, selected_config,
                    gain_frequency),
                  attributes_dict.get_configured_element(
                    'normalization_factor', channel_modif, selected_config,
                    None),
                  attributes_dict.get_configured_element(
                    'offset', channel_modif, selected_config, 0),
                  stage_id)
        return obj

    def __repr__(self):

        s = f'          PolesZeros(Poles={self.poles}, Zeros={self.zeros}, '
        s += f'Normalization Frequency={self.normalization_frequency:g}, '
        s += f'Normalization Factor={self.normalization_factor:g})'
        return s

    def calc_normalization_factor(self, stage_id=-1, debug=False):
        """
        Calculate the normalization factor for a given set of poles-zeros

        The norm factor A0 is calculated such that

        .. parsed-literal::
                                  sequence_product_over_n(s - zero_n)
                       A0 * abs(—--------------------------------------) == 1
                                  sequence_product_over_m(s - pole_m)

            for s_f = i*2pi*f if the transfer function is Laplace in radians
                      i*f if the transfer function is Laplace in Hertz

        There is no calculation for the digital z-transform

        Returns:
            normalization factor as a float or None if not Laplace
        """
        if not self.normalization_frequency:
            msg = ('No normalization frequency for PZ filter in '
                   f'stage #{stage_id}')
            logger.error(msg)
            raise ValueError(msg)

        A0 = 1.0 + (1j * 0.0)
        if self.transfer_function_type == "LAPLACE (HERTZ)":
            s = 1j * self.normalization_frequency
        elif self.transfer_function_type == "LAPLACE (RADIANS/SECOND)":
            s = 1j * 2 * m.pi * self.normalization_frequency
        else:
            msg = ("Don't know how to calculate normalization factor"
                   "for z-transform poles and zeros!")
            warnings.warn(msg)
            logger.warning(msg)
            return None

        for p in self.poles:
            A0 *= (s - p)
        for z in self.zeros:
            A0 /= (s - z)

        if debug:
            msg = f"poles={self.poles}, zeros={self.zeros}, s={s}, A0={A0}"
            print(msg)
            logger.debug(msg)

        A0 = abs(A0)
        return A0


class FIR(Filter):
    """
    FIR Filter

    Attributes:
        symmetry (str): filter symmetry, one of "EVEN", "ODD", "NONE"
        coefficients (list of floats)
        coefficient_divisor (float)
    """

    def __init__(self, filter_type, symmetry, coefficients,
                 coefficient_divisor, offset=0, stage_id=-1):

        """
        Constructor

        Args:
            filter_type (str): type of filter, necessarily 'FIR'
            symmetry (str): filter symmetry, one of "EVEN", "ODD", "NONE"
            coefficients (list): floats
            coefficient_divisor (float):
            offset (int): # of samples filter will offset an impulse
            stage_id (int): id of corresponding stage. Used for reporting only
        """
        self.symmetry = symmetry
        if symmetry not in ['ODD', 'EVEN', 'NONE']:
            msg = f'Illegal FIR symmetry: "{symmetry} in stage #{stage_id}"'
            logger.error(msg)
            raise TypeError()

        # Validate coefficients
        sum_coeff = 0
        coeff_cnt = 0
        for coeff in coefficients:
            sum_coeff += coeff
            coeff_cnt += 1
        if symmetry == 'EVEN':
            sum_coeff *= 2.
            coeff_cnt *= 2
        if symmetry == 'ODD':
            sum_coeff += sum_coeff - coefficients[-1]
            coeff_cnt += coeff_cnt - 1

        coeff = sum_coeff / coefficient_divisor
        coeff = round(coeff, 2)  # check up to two decimal places
        # last conditional verifies that there is at least one coeff
        if coeff != 1 and coeff != 0:
            msg = (f'Coefficient sum "{coeff}" not equal to one, num. '
                   f'coefficient: "{coeff_cnt}" in stage #{stage_id}')
            warnings.warn(msg)
            logger.warning(msg)

        self.coefficients = coefficients
        self.coefficient_divisor = coefficient_divisor

        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict,
                                  channel_modif={}, selected_config={},
                                  stage_id=-1):
        """
        Create FIR instance from an attributes_dict

        Args:
            filter_type (str): type of filter, necessarily 'FIR'
            attribute_list (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): the configuration
                description that will override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
        Returns:
            (:class:`.FIR`)
        """

        offset = attributes_dict.get_configured_element(
            'offset', channel_modif, selected_config, 0)
        if not offset:
            msg = 'No offset in FIR filter'
            logger.error(msg)
            raise TypeError(msg)

        obj = cls(filter_type,
                  attributes_dict.get_configured_element(
                    'symmetry', channel_modif, selected_config, None),
                  attributes_dict.get_configured_element(
                    'coefficients', channel_modif, selected_config, []),
                  attributes_dict.get_configured_element(
                    'coefficient_divisor', channel_modif, selected_config, 1.),
                  offset)
        return obj

    def __repr__(self):
        s = f'\tFIR("Symmetry={self.symmetry}",'
        s += f' Coefficients={self.coefficients},'
        s += f' Divisor={self.coefficient_divisor})'
        return s


class Coefficients(Filter):
    """
    Coefficients Filter Class

    Attributes:
        transfer_function_type (str): one of "ANALOG (RADIANS/SECOND)",
            "ANALOG (HERTZ)", or "DIGITAL"
        numerator_coefficients (list of floats)
        denominator_coefficients (list of floats)
    """

    def __init__(self, filter_type, transfer_function_type,
                 numerator_coefficients, denominator_coefficients, offset=0):
        """
        Constructor

        Args:
            filter_type (str): type of filter, necessarily 'Coefficients'
            transfer_function_type (str): one of "ANALOG (RADIANS/SECOND)",
                "ANALOG (HERTZ)", or "DIGITAL"
            numerator_coefficients (list): of floats
            denominator_coefficients (list): of floats
            offset (int): samples that an impulse is offset by the filter

        :param filter_type: type of filter, necessarily 'Coefficients'
        :type filter_type: str

        :param offset:
        :type offset: int
        """
        if transfer_function_type not in ["ANALOG (RADIANS/SECOND)",
                                          "ANALOG (HERTZ)",
                                          "DIGITAL"]:
            msg = f'Illegal transfer function type: "{transfer_function_type}"'
            logger.error(msg)
            raise TypeError(msg)

        self.transfer_function_type = transfer_function_type
        self.numerator_coefficients = numerator_coefficients
        self.denominator_coefficients = denominator_coefficients

        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict,
                                  channel_modif={},
                                  selected_config={}, stage_id=-1):
        """
        Create Coefficients instance from an attributes_dict

        Args:
            filter_type (str): type of filter, necessarily 'Coefficients'
            attribute_list (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): the configuration
                description that will override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
        Returns:
            (:class:`.Coefficients`)
        """
        obj = cls(filter_type,
                  attributes_dict.get_configured_element(
                    'transfer_function_type', channel_modif, selected_config,
                    'DIGITAL'),
                  attributes_dict.get_configured_element(
                    'numerator_coefficients', channel_modif, selected_config,
                    []),
                  attributes_dict.get_configured_element(
                    'denominator_coefficients', channel_modif, selected_config,
                    []),
                  attributes_dict.get_configured_element(
                    'offset', channel_modif, selected_config, 0)
                  )
        return obj

    def __repr__(self):
        s = '          Coefficients("Transfer Function Type='
        s += f'{self.transfer_function_type}", '
        s += f'Numerator={self.numerator_coefficients}, '
        s += f'Denominator={self.denominator_coefficients})'
        return s


class ResponseList(Filter):
    """
    ResponseList Filter

    Attributes:
        response_list (list of floats): list of responses instead of function
            coefficients
    """

    def __init__(self, filter_type, response_list, offset=0):
        """
        Constructor

        Args"
            filter_type (str): type of filter, necessarily 'ResponseList'
            response_list (list): floats specifying response
            offset (int): samples that an impulse is offset by the filter
        """
        self.response_list = response_list
        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict,
                                  channel_modif={},
                                  selected_config={}, stage_id=-1):
        """
        Create Response List instance from an attributes_dict

        Args:
            filter_type (str): type of filter, necessarily 'ResponseList'
            attribute_list (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): the configuration
                description that will override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
        Returns:
            (:class:`.ResponseList`)
        """
        obj = cls(filter_type,
                  attributes_dict.get_configured_element(
                    'elements', channel_modif, selected_config, []),
                  attributes_dict.get_configured_element(
                    'offset', channel_modif, selected_config, 0))
        return obj

    def __repr__(self):
        return f'          ResponseList("{self.response_list}")'


class Analog(PolesZeros):
    """
    Analog Filter (Flat PolesZeros filter)
    """

    def __init__(self, filter_type, offset=0, normalization_frequency=1.):
        """
        Constructor

        Args:
            filter_type (str): necessarily 'Analog'
            offset (int):
            normalization_frequency (float):
        :type offset: int
        """
        # self.poles = []
        # self.zeros = []
        # self.normalization_frequency = 1.
        # self.normalization_factor = None
        # super().__init__(filter_type, "LAPLACE (RADIANS/SECOND)",
        #                  self.poles, self.zeros, self.normalization_frequency,
        #                  self.normalization_factor, offset)
        super().__init__(filter_type, "LAPLACE (RADIANS/SECOND)",
                         [], [], normalization_frequency, None, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict,
                                  channel_modif={},
                                  selected_config={}, stage_id=-1,
                                  gain_frequency=1.):
        """
        Create Analog instance from an attributes_dict

        Args:
            filter_type (str): type of filter, necessarily 'Analog'
            attribute_list (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): the configuration
                description that will override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
            gain_frequency (float): frequency at which gain was specified.
        Returns:
            (:class:`.Analog`)
        """
        obj = cls(filter_type,
                  attributes_dict.get_configured_element(
                    'offset', channel_modif, selected_config, 0),
                  gain_frequency)
        return obj

    def __repr__(self):
        return '          Analog()'


class Digital(Coefficients):
    """
    Digital Filter (Flat Coefficients filter)
    """

    def __init__(self, filter_type, offset=0):
        """
        Constructor

        :param filter_type: type of filter, necessarily 'Digital'
        :type filter_type: str
        :param offset:
        :type offset: int
        """
        self.transfer_function_type = 'DIGITAL'
        self.numerator_coefficients = [1.0]
        self.denominator_coefficients = []

        super().__init__(filter_type, "DIGITAL", self.numerator_coefficients,
                         self.denominator_coefficients, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict,
                                  channel_modif={},
                                  selected_config={}, stage_id=-1):
        """
        Create Digital instance from an attributes_dict

        Args:
            filter_type (str): type of filter, necessarily 'Digital'
            attribute_list (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): the configuration
                description that will override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
        Returns:
            (:class:`.Digital`)
        """

        obj = cls(filter_type, attributes_dict.get_configured_element(
            'offset', channel_modif, selected_config, 0))
        return obj

    def __repr__(self):
        return '          Digital()'


class AD_Conversion(Coefficients):
    """
    AD_Conversion Filter (Flat Coefficients filter)

    Attributes:
        input_full_scale (float)
        output_full_scale (float)
    """

    def __init__(self, filter_type, input_full_scale, output_full_scale,
                 offset=0):
        """
        Constructor

        Args:
            filter_type (str): type of filter, necessarily 'ADConversion'
            input_full_scale (float): A/D's input full scale (volts)
            output_full_scale (float): corresponding output full scale (counts)
            offset (int): number of samples that the filter offsets an
                impulse
        """

        self.transfer_function_type = 'DIGITAL'
        self.numerator_coefficients = [1.0]
        self.denominator_coefficients = []
        self.input_full_scale = input_full_scale
        self.output_full_scale = output_full_scale

        super().__init__(filter_type, "DIGITAL",
                         self.numerator_coefficients,
                         self.denominator_coefficients, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict,
                                  channel_modif={},  selected_config={},
                                  stage_id=-1):
        """
        Create AD_Conversion instance from an attributes_dict

        Args:
            filter_type (str): type of filter, necessarily 'ADConversion'
            attribute_list (dict or list of dicts): information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): configuration
                description to override or complement default values
            stage_id (int): id of corresponding stage. Used for reporting only
        Returns:
            (:class:`.ADConversion`)
        """
        obj = cls(filter_type,
                  attributes_dict.get_configured_element(
                    'input_full_scale', channel_modif, selected_config, None),
                  attributes_dict.get_configured_element(
                    'output_full_scale', channel_modif, selected_config, None),
                  attributes_dict.get_configured_element(
                    'offset', channel_modif, selected_config, 0))
        return obj

    def __repr__(self):
        s = f'\tADConversion(Input Full Scale={self.input_full_scale:g}, '
        s += f'Output Full Scale={self.output_full_scale})'
        return s



"""
Write bash scripts to convert LCHEAPO data to SeisComp Data Structure (SDS)
  - Includes clock drift and leap-second correction (must be entered in obsinfo
    network files)
  - For onboard processing, does not apply drift at the record level, nor set
    associated record header flags or fill in record header time_correction
  - Scripts require the lcheapo python package, and only work in a BASH shell
"""
import os.path
from pathlib import Path
import warnings

import obsinfo
from obsinfo.network import Network
from ..misc.discoveryfiles import (Datapath)
from ..obsMetadata.obsmetadata import (ObsMetadata)
from .LCHEAPO import _get_ref_code

SEPARATOR_LINE = "\n# " + 60 * "=" + "\n"


def process_script(network_code, station, station_dir, input_dir=".",
                   output_dir="../", include_header=True):
    """
    Writes script to transform raw OBS data to SeisComp Data Structure

    Arguments:
        network_code (str): FDSN network_code
        station (:class:`.Station`): the station to process
        station_dir (str): the base directory for the station data
        input_dir (str): directory beneath station_dir for LCHEAPO data
        output_dir (str): directory beneath station_dir for SDS directory
        include_header (bool): include the header that sets up paths
                               (should be done once)
    """
    fixed_dir = "lcheapo_fixed"
    s = ""
    if include_header:
        s += _header(station.label)
    s += _setup_variables(station_dir)
    s += _lcfix_commands(station, input_dir, fixed_dir)
    s += _lc2sds_commands(network_code, station, fixed_dir, output_dir)
    s += f'# Remove intermediate files\n'
    s += f'command rm -r {station_dir}/{fixed_dir}\n'
    # s += _force_quality_commands(output_dir, "D")
    return s


def _header(station_name):

    s = "#!/bin/bash\n"
    # s += SEPARATOR_LINE
    s += f'echo "Working on station {station_name}"'
    # s += SEPARATOR_LINE
    return s


def _setup_variables(station_dir):
    """
    station_dir: base directory for station data files
    """

    s = SEPARATOR_LINE + "# LC2SDS STEPS\n"
    s += "#  - Set up paths\n"
    s += f"STATION_DIR={station_dir}\n"
    s += f"\n"
    return s


def _lcfix_commands(station, in_path, out_path, in_fnames="*.raw.lch"):

    """
        Write an lcfix command line

        Inputs:
            in_path:       relative path to directory containing input files
            out_path:      relative path to directory for output files
            in_fnames:     search string for input files within in_path
         Output:
            string of bash script lines
    """

    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running LCFIX: Fix common LCHEAPO data bugs"\n'
    # s += f'echo "{"-"*60}"\n'
    # s += f'in_dir="{in_path}"\n'
    # s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory and collect input filenames\n"
    # s += "mkdir $STATION_DIR/$out_dir\n"
    s += f"mkdir $STATION_DIR/{out_path}\n"
    # s += "# - Collect input filenames\n"
    s += f"command cd $STATION_DIR/{in_path}\n"
    s += f"lchfiles=$(command ls {in_fnames})\n"
    s += "command cd -\n"
    s += 'echo "lchfile(s): " $lchfiles\n'

    s += "# - Run executable\n"
    # s += 'lcfix $lchfiles -d "$STATION_DIR" -i $in_dir -o $out_dir\n'
    s += f'lcfix $lchfiles -d "$STATION_DIR" -i "{in_path}" -o "{out_path}"\n'
    s += "\n"

    return s


def _lc2sds_commands(network_code, station, in_path, out_path,
                     in_fnames="*.fix.lch",
                     out_fnames_model="%E.%S.00.%C.%Y.%D.%T.mseed",
                     force_quality_D=True):
    """
    Write an lc2sds command line

    Arguments:
        network_code (str): network code
        station (:class:`Station`): station to process
        in_path (str): relative path to directory containing input files
        in_fnames (str): search string for input files within in_path
        out_path (str): relative path to directory for output files
        out_fnames_model (str): model for output filenames (should change
            default to '%E.%S.%L.%C.%Y.%D.%T.mseed' once lc2ms handles
            location codes)
        force_quality_D: make a separate call to msmod to force the data
            quality to "D" (should be unecessary once lc2ms is upgraded)
    Returns:
        (str): bash script lines
    """

    network_code = network_code
    station_code = station.label
    obs_type = _get_ref_code(station.instrumentation)
    # obs_SN = station.instrumentation.equipment.serial_number
    # channel_corresp = station.instrument.channel_correspondances()

    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running LC2SDS_weak: Transform LCHEAPO data to SeisComp Data Structure"\n'
    # s += f'echo "{"-"*60}"\n'
    # s += f'in_dir="{in_path}"\n'
    # s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory and collect input filenames\n"
    # s += "mkdir -p $STATION_DIR/$out_dir\n"
    s += f"mkdir -p $STATION_DIR/{out_path}\n"

    # s += "# - Collect input filenames\n"
    # s += "command cd $STATION_DIR/$in_dir\n"
    s += f"command cd $STATION_DIR/{in_path}\n"
    s += f"lchfiles=$(command ls {in_fnames})\n"
    s += "command cd -\n"
    s += 'echo "lchfile(s): " $lchfiles\n'

    s += "# - Run executable\n"
    # s += 'lc2SDS_weak $lchfiles -d "$STATION_DIR" -i $in_dir -o $out_dir '
    s += f'lc2SDS_weak $lchfiles -d "$STATION_DIR" -i "{in_path}" -o "{out_path}" '
    s += f'--network "{network_code}" '
    s += f'--station "{station_code}" '
    s += f'--obs_type "{obs_type}" '
    leaptimes, leaptypes = [], []
    ccld = None
    for proc in station.processing.attributes:
        if 'clock_correction_linear' in proc:
            if ccld is not None:
                warnings.warn('more than one clock_correction_linear, '
                              'only applying first')
            else:
                ccld = proc['clock_correction_linear']
        elif 'clock_correction_leapsecond' in proc:
            leaptimes.append(proc['clock_correction_leapsecond']['time'])
            leaptypes.append(proc['clock_correction_leapsecond']['type'])
    if ccld is not None:
        s += f'--start_times "{ccld["start_sync_reference"]}" '
        if ccld["start_sync_instrument"]:
            s += f' "{ccld["start_sync_instrument"]}" '
        s += f'--end_times "{ccld["end_sync_reference"]}" "{ccld["end_sync_instrument"]}" '
    if leaptimes:
        s += f'--leapsecond_times "{" ".join(leaptimes)}"'
        s += f'--leapsecond_types "{" ".join(leaptypes)}"'
    s += "\n\n"

    return s


def _console_script(argv=None):
    """
    Create a bash-script to convert LCHEAPO data to SDS, with time correction
    """
    from argparse import ArgumentParser, RawDescriptionHelpFormatter

    parser = ArgumentParser(prog="obsinfo-makescripts_LS2SDS",
                            description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument("network_file", help="Network information file")
    parser.add_argument("station_data_path",
                        help="Base path containing the station directories")
    parser.add_argument("-i", "--input_dir", default=".",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "containing input *.raw.lch files "
                             "(default: %(default)s)")
    parser.add_argument("-o", "--output_dir", default="../",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "to put output SDS directory "
                             "(default: %(default)s)")
    parser.add_argument("--suffix", default="_LC2SDS",
                        help="suffix for script filename "
                             "(default: %(default)s)")
    parser.add_argument("--append", action="store_true",
                        help="append to existing script file")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="increase output verbosity")
    parser.add_argument("--no_header", action="store_true",
                        help="do not include a script header")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="run silently")
    args = parser.parse_args()

    if not args.quiet:
        print("Creating LC2SDS_weak process scripts, ", end="", flush=True)

    # READ IN NETWORK INFORMATION
    dp = Datapath()
    if args.verbose:
        print(f'Reading network file: {args.network_file}')
    args.network_file = str(Path(os.getcwd()).joinpath(args.network_file))
    info_dict = ObsMetadata.read_info_file(args.network_file, dp, False)
    net_dict = info_dict.get('network', None)
    if not net_dict:
        return
    if args.verbose:
        print(f'Processing network file: {args.network_file}')
    network = Network(ObsMetadata(net_dict))
    # network = oi_Network(args.network_file)

    if not args.quiet:
        print(f"network {network.fdsn_code}, stations ", end="", flush=True)
        if args.verbose:
            print("")

    first_time = True
    for station in network.stations:
        if not args.quiet:
            if args.verbose:
                print(f"\t{station.label}", end="")
            else:
                if first_time:
                    print(f"{station.label}", end="", flush=True)
                else:
                    print(f", {station.label}", end="", flush=True)
        station_dir = os.path.join(args.station_data_path, station.label)
        script = process_script(
            network.fdsn_code,
            station,
            station_dir,
            # args.distrib_path,
            input_dir=args.input_dir,
            output_dir=args.output_dir,
            include_header=not args.no_header,
        )
        fname = "process_" + station.label + args.suffix + ".sh"
        if args.verbose:
            print(f" ... writing file {fname}", flush=True)
        if args.append:
            write_mode = "a"
        else:
            write_mode = "w"
        with open(fname, write_mode) as f:
            # f.write('#'+'='*60 + '\n')
            f.write(script)
            f.close()
        first_time = False
    if not args.verbose and not args.quiet:
        print("")



"""
Instrument and Operator classes
"""
# Standard library modules
import warnings
import re
import logging

from obspy.core.inventory.util import (Operator as obspy_Operator,
                                       Person, PhoneNumber)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Operator(object):  # Previously a Channel subclass, but don't see why
    """
    Contact information for the operator of the instrumentation or network

    Attributes:
        reference_name (str): Reference name of operator as known to the FDSN
        full_name (str): Full name of operator
        contact (dict of two strings: first_name and last_name)
        email (str)
        country_code (str)
        area_code (str): unused for non-American phones.
        phone_number (str): formatted according to strange FDSN rules
            ([0-9]+-[0-9]
        website (str)
    """

    def __init__(self, attributes_dict):
        """
        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): operator information
        """
        if not attributes_dict:
            raise ValueError('No attributes dict!')

        self.reference_name = attributes_dict.get('reference_name', None)
        self.full_name = attributes_dict.get('full_name', '')
        contact_name = attributes_dict.get('contact', None)

        if contact_name:
            self.contact = contact_name.get('first_name', '') \
                           + " " + contact_name.get('last_name', '')
        else:
            self.contact = ""

        self.email = attributes_dict.get('email', None)
        self.country_code, self.area_code, self.phone_number = \
            self.convert_phone_number(attributes_dict.get('phone_number',
                                                          None))
        self.website = attributes_dict.get('website', None)

    def __str__(self):
        s = f'\nOperator: Reference Name={self.reference_name}\n'
        s += f'\tFull Name={self.full_name}\n'
        if self.contact:
            s += f'\tContact Name={self.contact}\n'
        s += f'\tEmail={self.email}\n'
        s += f'\tPhone_number={self.phone_number}\n'
        s += f'\tWebsite={self.website}'

        return s

    def convert_phone_number(self, phone):
        """
        Try to convert international numbers to the FDSN American standard.
        If already in American standard, use area code.

        Args:
            phone (str): phone number in (hopefully) one of several
                recognisable formats
        Returns:
            (tuple):
                ``country_code``
                ``area_code``: default=0
                ``phone_number``
        """
        country_code = None
        area_code = "0"
        phone_number = ""

        if not phone:
            return (country_code, area_code, phone_number)

        if isinstance(phone, dict):
            country_code = phone.get('country_code', None)
            area_code = phone.get('area_code', "")
            phone_number = phone.get('phone_number', "")
        elif isinstance(phone, type("")):
            # For reference:
            # country = re.compile("^(\+{0,1}|00)[0-9]{1,3}$")
            # area = re.compile("^\({0,1}[0-9]{3}\){0,1}$")
            # phone = re.compile("^[0-9]{3}\-[0-9]{4}$")

            us_phone_ptrn = '(?P<country>(\\+{0,1}|00)[0-9]{1,3}) '\
                            '*(?P<area>\\({0,1}[0-9]{3}\\){0,1}) '\
                            '*(?P<phone>[0-9]{3}\\-[0-9]{4})$'
            us_phone_re = re.compile(us_phone_ptrn)

            us_phone_match = us_phone_re.match(phone)
            if us_phone_match:
                country_code = us_phone_match.group['country']
                area_code = us_phone_match.group['area']
                phone_number = us_phone_match.group['phone']
            else:
                c_code_plus_ptn = '^\+([0-9]{1,3})'
                c_code_zero_ptn = "^00([0-9]{1,3})"
                c_code_plus_re = re.compile(c_code_plus_ptn)
                c_code_zero_re = re.compile(c_code_zero_ptn)
                phone_ptn = "(?P<phone>(\([0-9]+\))* *([0-9]+[ \-\.]*)*[0-9]+)$"

                c_code = c_code_plus_re.match(phone)
                if not c_code:  # | for alternatives in regex doesn't work
                    c_code = c_code_zero_re.match(phone)
                    phone_re = re.compile(c_code_zero_ptn + " *" + phone_ptn)
                else:
                    phone_re = re.compile(c_code_plus_ptn + " *" + phone_ptn)

                if c_code:
                    country_code = c_code.group(1)

                mtch = phone_re.match(phone)
                if mtch:
                    phone_number = mtch.group('phone')
                    # The following is done to avoid FDSN reg exp restrictions
                    # for phones, American based
                    for chr in ["(", ")", ".", "-", " "]:
                        phone_number = phone_number.replace(chr, "")
                    phone_number = phone_number[0:3] + "-" + phone_number[3:]
        else:
            pass
        return (country_code, area_code, phone_number)

    def to_obspy(self):
        emails = [self.email] if self.email else []
        phones = [PhoneNumber(self.area_code,
                              self.phone_number,
                              self.country_code)]\
            if self.phone_number else []
        names = [self.contact] if self.contact else []
        agencies = [self.reference_name] \
            if self.reference_name else []
        person = Person(names=names,
                        agencies=agencies,
                        emails=emails,
                        phones=phones)
        return obspy_Operator(agency=self.full_name,
                              contacts=[person],
                              website=self.website)
    
        



"""
ResponseStages and Stage classes
"""
# Standard library modules
import warnings
import re
import logging

# Non-standard modules
from obspy.core.inventory.response import (PolesZerosResponseStage,
                                           FIRResponseStage,
                                           CoefficientsTypeResponseStage,
                                           ResponseListResponseStage)
# from obspy.core.inventory.response import Response as obspy_Response
import obspy.core.util.obspy_types as obspy_types

# Local modules
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)
from .filter import (Filter, PolesZeros, FIR, Coefficients, ResponseList,
                     Analog, Digital, AD_Conversion)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class ResponseStages():
    """
    An ordered list of :class:`Stage`, representing a total response.

    Attributes:
        stages (list of objects of :class:`Stage`)
        obspy_stages (list of objects of different obspy classes according
            to the type of stage)
    """
    def __init__(self, attribute_list, channel_modif={}, selected_config={},
                 delay_correction=None):
        """
        Constructor

        Method can be invoked with attribute_list as a single element of class
        Stage or as a list of them.

        Args:
            attribute_list (list of dicts or dict): list of information file
                dictionaries for stages
            channel_modif (dict or :class:`.ObsMetadata`): channel
                modifications inherited from station
            selected_config (dict or :class:`.ObsMetadata`): configuration
                description to override or complement default values
            delay_correction (float): used only for datalogger: the delay
                correction for the entire instrument
        """

        if attribute_list is None:
            msg = 'No stages found in information file'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)
        elif not isinstance(attribute_list, list):  # Single element
            self.stages = Stage(ObsMetadata(attribute_list),
                                channel_modif,
                                selected_config,
                                delay_correction,
                                1)
        else:
            self.stages = []
            for s, i in zip(attribute_list, range(0, len(attribute_list))):
                if delay_correction is None:
                    correction = None
                elif i == len(attribute_list)-1:
                    correction = delay_correction
                else:
                    correction = 0
                self.stages.append(Stage(ObsMetadata(s),
                                         channel_modif,
                                         selected_config,
                                         correction,
                                         i+1))

        self.obspy_stages = self.to_obspy()
        # delay, delay corrections and sample rates (if not present
        # for digital stages)  will be calculated for the whole instrumentation

    def to_obspy(self):
        """
        Return list of obspy stage classes

        Each class will have a different obspy class

        :returns: list of objects of obspy classes
        """
        if not self.stages:
            return None
        obspy_stages = [s.to_obspy() for s in self.stages]
        return obspy_stages

    def __repr__(self):

        if self.stages is not None:
            return f' Number of response stages: {len(self.stages):d}'
        else:
            return 'Response_Stages: (empty)'


class Stage(object):
    """
    Stage is a discrete portion of the response of the instrument

    Attributes:
        name (str): name of stage, if any
        description (str): description of stage
        input_units (str): validated in schema
        input_units_description (str)
        output_units (str): validated in schema
        output_units_description (str)
        gain (float): value of gain
        gain_frequency (float): frequency at which gain is measured
        filter (object of :class:`Filter`)
        stage_sequence_number: sequence number in total response, assigned
            later
        input_sample_rate (float): input sample rate in sps
        delay (float): delay in seconds of stage. If not present, will be
            calculated from offset from digital stages
        decimation_factor (float): decimation factor of stage
        correction (float) : delay correction. Calculated from instrument
            delay correction
        polarity (str, either "+" or "-"): whether stage changes polarity
        calibration_date (str in date format): calibration date of stage
        instrument_sensitivity (float): Not used, set to None. Sensitivity
            is calculated for the whole response.

    """
    def __init__(self, attributes_dict, channel_modif_list={},
                 selected_config={},  correction=None,
                 sequence_number=-1):
        """
        Args:
            attributes_dict (dict or :class:`ObsMetadata`): attributes of
                component
            channel_modif (dict or :class:`ObsMetadata`): channel modifications
                inherited from station
            selected_config (dict or :class:`ObsMetadata`): the configuration
                description that will override or complement default values
            correction (float): used only for datalogger, it's the delay
                correction of the whole instrument
            sequence_number (int): sequence number, starting at 1. First
                assigned within component, then for the whole instrument
                response
        """
        if attributes_dict is None:
            return None

        channel_modif = self.get_selected_modifications(
            channel_modif_list, str(sequence_number-1))

        name = attributes_dict.get_configured_element(
            'name', channel_modif, selected_config, '')

        self.name = name if name else None
        # stage id is either the name or None

        self.description = attributes_dict.get_configured_element(
            'description', channel_modif, selected_config, '')
        self.input_units = attributes_dict.get_configured_element(
            'input_units', channel_modif, selected_config, None).get(
            'name', None)

        self.output_units = attributes_dict.get_configured_element(
            'output_units', channel_modif, selected_config, None).get(
            'name', None)

        gain_dict = ObsMetadata(attributes_dict.get('gain', {}))
        if gain_dict:
            modif_gain_dict = channel_modif.get('gain', {})
            self.gain = gain_dict.get_configured_element(
                'value', modif_gain_dict, selected_config, 1.0)
            self.gain_frequency = gain_dict.get_configured_element(
                'frequency', modif_gain_dict, selected_config, 0.0)
        else:
            msg = f'No gain specified in stage {self.name}'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        # filter cannot be changed by configuration, but it can by channel
        # modification
        self.filter = Filter.dynamic_class_constructor(
            ObsMetadata(attributes_dict.get('filter', None)),
            channel_modif.get('filter', None),
            selected_config, self.name, self.gain_frequency)
        if not self.filter:
            msg = f'No filter in stage {self.name}'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        self.stage_sequence_number = sequence_number

        input_dict = attributes_dict.get_configured_element(
                'input_units', channel_modif, selected_config, None)
        if input_dict:
            self.input_units_description = input_dict.get('description', None)
        output_dict = attributes_dict.get_configured_element(
                'output_units', channel_modif, selected_config, None)
        if output_dict:
            self.output_units_description = output_dict.get('description',
                                                            None)
        self.input_sample_rate = attributes_dict.get_configured_element(
            'input_sample_rate', channel_modif, selected_config, None)

        # Set an unconfigured delay to None so that it can be
        # calculated in self.calculate_delay(), once input_sample_rate
        # is known
        self.delay = attributes_dict.get_configured_element(
            'delay', channel_modif, selected_config, None)
        self.decimation_factor = attributes_dict.get_configured_element(
            'decimation_factor', channel_modif, selected_config, 1)
        self.correction = correction

        # default polarity is positive
        self.polarity = attributes_dict.get_configured_element(
            'polarity', channel_modif, selected_config, 1)

        self.calibration_date = attributes_dict.get_configured_element(
            'calibration_date', selected_config, None)

        self.instrument_sensitivity = None
        # Overall sensitivity will be calculated using obspy and stored in
        # first stage

    @property
    def output_sample_rate(self):
        """
        Output sample rate is not specified but calculated from
        input sample rate and decimation factor
        """
        if self.input_sample_rate and self.decimation_factor:
            return self.input_sample_rate / self.decimation_factor
        else:
            return None

    def calculate_delay(self):
        """
        Calculates delay

        Delay is a function of filter offset for digital filters if not
        specified in info file
        """

        delay = self.delay
        offset = self.filter.offset
        input_sample_rate = self.input_sample_rate

        if delay is None:
            if offset is not None:
                if not input_sample_rate:
                    # Delay is already none, leave it like that
                    msg = 'Stage delay is impossible to calculate out of '\
                          'filter offset with an Unspecified input sample '\
                          f'rate in stage {self.name}. Setting to "None"'
                    logger.warning(msg)
                    warnings.warn(msg)
                else:
                    self.delay = offset / input_sample_rate
            else:
                self.delay = None

        else:
            self.delay = delay

    def get_selected_modifications(self, modif_dict, key_code):
        """
        Select which channel modifications specified at station level apply
        to a given stage,  with the stage number (WITHIN an instrument
        component) as key code

        Args:
            modif_dict (dict or :class:`.ObsMetadata`): channel modifications
                inherited from station
            key_code (str): key to the selected configuration using special
                regex format, e.g.: "*", "[1,2,4]" or "[1-3]"
        """
        default_dict = range_dict = {}
        modif = modif_dict.get(key_code, {})

        overlap = False
        for k, v in modif_dict.items():
            if k[0] == "*":
                default_dict = v
            elif k[0][0] == "[":
                if re.match(k, key_code):
                    if overlap:
                        msg = 'There is an overlap in response stage '\
                              'modifications. Taking the first applicable '\
                              'pattern.'
                        warnings.warn(msg)
                        logger.warning(msg)
                        break  # Will only use first match, to avoid conflicts
                    range_dict = v
                    overlap = True

        # Gather all modifications in a single dict
        # Do this in order: particular mods have priority over range
        # specific which has priority over default
        for k, v in range_dict.items():
            if k not in modif:
                modif[k] = v

        for k, v in default_dict.items():
            if k not in modif:
                modif[k] = v

        return modif

    def __repr__(self):
        s = f'\n     Stage("{self.name}", "{self.description}", '
        s += f'"{self.input_units}", "{self.output_units}", '
        s += f'{self.gain}, {self.gain_frequency:g}, '
        s += f'{type(self.filter)}'
        if not self.stage_sequence_number == -1:
            s += f', stage_sequence_number="{self.stage_sequence_number}"'
        if self.input_units_description:
            s += f', input_units_description="{self.input_units_description}"'
        if self.output_units_description:
            s += ', output_units_description='
            s += f'"{self.output_units_description}"'
        if self.input_sample_rate:
            s += f', input_sample_rate={self.input_sample_rate}'

        s += f', decimation_factor={self.decimation_factor}'
        s += f', delay={self.delay}'
        s += f', correction={self.correction}'

        if self.calibration_date:
            s += f', calibration_dates={self.calibration_date}'
        s += ')'
        return s

    def to_obspy(self):
        """
        Return equivalent *obspy.core.inventory.response* classes stages:

        Possible stage classes:

           * PolesZerosResponseStage
           * FIRResponseStage
           * CoefficientsTypeResponseStage
           * ResponseListResponseStage
           * Response

        :returns: object of one the above classes
        """
        filt = self.filter
        # Invariable position arguments for all
        args = (self.stage_sequence_number, self.gain, self.gain_frequency,
                self.input_units, self.output_units)

        if (isinstance(filt, PolesZeros) or isinstance(filt, Analog)):
            if not filt.normalization_frequency:
                filt.normalization_frequency = self.gain_frequency

            if filt.type == "PolesZeros":
                # Only iterate if zeros and poles exist
                PZ_zeros = [obspy_types.ComplexWithUncertainties(t)
                            for t in filt.zeros]
                PZ_poles = [obspy_types.ComplexWithUncertainties(t)
                            for t in filt.poles]
            else:
                PZ_zeros = filt.zeros
                PZ_poles = filt.poles

            obspy_stage = PolesZerosResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,
                # PolesZeros-specific
                pz_transfer_function_type=filt.transfer_function_type,
                normalization_frequency=filt.normalization_frequency,
                zeros=PZ_zeros,
                poles=PZ_poles,
                normalization_factor=filt.normalization_factor)

        elif isinstance(filt, FIR):
            obspy_stage = FIRResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,
                # FIR-specific
                symmetry=filt.symmetry,
                coefficients=[obspy_types.FloatWithUncertaintiesAndUnit(
                    c / filt.coefficient_divisor) for c in filt.coefficients])

        elif (isinstance(filt, Coefficients) or isinstance(filt, Digital)
              or isinstance(filt, AD_Conversion)):
            if filt.type == "Coefficients":
                # Only iterate if zeros and poles exist
                c_numerator = [obspy_types.FloatWithUncertaintiesAndUnit(
                    n, lower_uncertainty=0.0, upper_uncertainty=0.0)
                    for n in filt.numerator_coefficients]
                c_denominator = [obspy_types.FloatWithUncertaintiesAndUnit(
                    n, lower_uncertainty=0.0, upper_uncertainty=0.0)
                    for n in filt.denominator_coefficients]
            else:
                c_numerator = filt.numerator_coefficients
                c_denominator = filt.denominator_coefficients

            obspy_stage = CoefficientsTypeResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,
                # CF-specific
                cf_transfer_function_type=filt.transfer_function_type,
                numerator=c_numerator,
                denominator=c_denominator
                )

        elif isinstance(filt, ResponseList):
            obspy_stage = ResponseListResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,
                # ResponeList-specific
                response_list_elements=filt.response_list)

        else:
            msg = 'Unhandled response stage type in stage '\
                  f'#{self.stage_sequence_number}: "{filt.type}"'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        return obspy_stage



"""
Instrumentation class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules

# obsinfo modules
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)
from .instrument_component import Equipment
from .channel import Channel
# from .operator_class import Operator

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Instrumentation(object):
    """
    One or more Instruments. Part of an obspy/StationXML Station

    Methods convert info files to an instance of this class. No equivalent
    obspy/StationXML class

    A more detailed description the class and its attributes is found in XXX

    Attributes:
        equipment (:class:`Equipment`):
        channels (list): list of channels (:class:`Channel`)
    """
    def __init__(self, attributes_dict_or_list, locations,
                 start_date, end_date, channel_modifications={},
                 serial_number=None):
        """
        Constructor

        attributes_dict may contain a configuration_selection for the
        instrumentation and the corresponding configs for the components:
        datalogger, preamplifier and sensor

        Args:
            attributes_dict_or_list (dict/ObsMetadata or list of same):
                instrumentation(s) attributes
            locations (list):  of :class:`Locations`
            start_date (str): start date
            end_date (str): end date
            channel_modifications (dict or :class:`ObsMetadata`):
                modification of attributes per channel specified in stations
            serial_number (str): instrumentation (OBS) serial number.
                Configures nothing for now

        locations, start_date and end_date are inherited from the
        corresponding attributes in a station to fill out StationXML fields.
        It is assumed an instrumentation default location,
        start date and end_date are the same as its station's.
        """
        # Syntax checking - Check whether
        if not attributes_dict_or_list:
            msg = 'No instrumentation attributes'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)
        elif isinstance(attributes_dict_or_list, list):
            # Easy mistake of including a dash in the yaml file
            msg = 'Instrumentation should be unique, not a list. Removing '\
                  'list embedding.'
            warnings.warn(msg)
            logger.warning(msg)

            if len(attributes_dict_or_list) > 0:
                attributes_dict = attributes_dict_or_list[0]
            else:
                msg = 'No instrumentation attributes'
                warnings.warn(msg)
                logger.error(msg)
                raise TypeError(msg)
        else:
            attributes_dict = attributes_dict_or_list

        # self.operator = Operator(attributes_dict.get('operator', None))
        self.equipment = Equipment(
            ObsMetadata(attributes_dict.get('equipment', None)), {}, {})
        if serial_number is not None:
            self.equipment.serial_number = serial_number
        self.equipment.obspy_equipment = self.equipment.to_obspy()

        das_channels = attributes_dict.get('channels', {})
        channel_default = das_channels.get('default', {})
        if channel_default:
            del das_channels['default']

        # v here is the attributes_dict of each das_channel
        self.channels = [Channel(k, v, locations, start_date, end_date,
                                 self.equipment.obspy_equipment,
                                 channel_default, channel_modifications)
                         for k, v in das_channels.items()]

    def __repr__(self):
        s = f'\n\nInstrumentation({len(self.channels)-1} channels)'
        return s



"""
Station Class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.station import Station as obspy_Station

from obspy.core.inventory.util import (Site, Comment)
from obspy.core.utcdatetime import UTCDateTime
# from obspy.taup.seismic_phase import self_tokenizer

# obsinfo modules
from .processing import Processing
from obsinfo.instrumentation import (Instrumentation, Location)
from .operator_class import Operator
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Station(object):
    """
    Station. Equivalent to obspy/StationXML Station

    Methods convert info files to an instance of this class and convert the
    object to an `obspy` object.

    Attributes:
        label:
        site (str):
        start_date (str with date format): station start date
        end_date (str with date format): station end date
        location_code (str):
        restricted_status (str): status of station
        locations (list of :class:`.Location`)
        location (:class:`.Location`): default location code
            of channels, corresònding to `location_code`
        instrumentation (:class:`.Instrumentation` or list of
            :class:`.Instrumentation`):
        processing (list of objects of :class:`.Processing`) : attributes
            for clock correction processing
        comments (list of str):
        extras (list of str):
        obspy_station (:class:`obspy.core.inventory.station.Station`):
            Equivalent obspy object
    """

    def __init__(self, label, attributes_dict, station_only=False,
                 stations_operator=None):
        """
        Constructor

        Args:
            attributes_dict: (dict or :class:`.ObsMetadata`): dictionary
                from station or network info file with YAML or JSON attributes
            station_only (boolean): Creates object with no instrumentation
            stations_operator (:class:`.Operator`): default station operator
        Raises:
            TypeError
        """
        if not attributes_dict:
            msg = 'No station attributes'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        self.label = label
        self.site = attributes_dict.get("site", None)

        start_date = ObsMetadata.validate_date(
            attributes_dict.get("start_date", None))
        self.start_date = UTCDateTime(start_date) if start_date else None
        end_date = ObsMetadata.validate_date(attributes_dict.get("end_date",
                                                                 None))
        self.end_date = UTCDateTime(end_date) if end_date else None
        self.location_code = attributes_dict.get("location_code", None)
        self.serial_number = attributes_dict.get("serial_number", None)
        self.restricted_status = attributes_dict.get("restricted_status", None)
        op = attributes_dict.get("operator", None)
        if op is None:
            self.operator = stations_operator
        else:
            self.operator = Operator(op)

        self.locations = {c: Location(v) for c, v in
                          attributes_dict.get('locations', None).items()}
        self.location = Location.get_location_from_code(
            self.locations, self.location_code, "station", self.label)

        instr_list = attributes_dict.get('instrumentations', None)
        if instr_list and isinstance(instr_list, list):
            channel_mods = attributes_dict.get('channel_modifications', {})
            self.instrumentation = [Instrumentation(inst, self.locations,
                                                    start_date, end_date,
                                                    channel_mods,
                                                    self.serial_number)
                                    for inst in instr_list]
        else:
            instr_dict = attributes_dict.get('instrumentation', None)
            channel_mods = attributes_dict.get('channel_modifications', {})
            # print()
            # print(f'Station: {attributes_dict=}, {channel_mods=}')

            if instr_dict:
                self.instrumentation = Instrumentation(
                    instr_dict, self.locations, start_date, end_date,
                    channel_mods, self.serial_number)
            elif station_only:
                # Create StationXML file with no instrumentation w/o
                # raising an exception
                self.instrumentation = None
            else:
                msg = f'No instrumentation in station "{self.site}"'
                # warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)

        # Locations, start_date and end_date are used to creat obspy Channel
        self.comments = attributes_dict.get("commnents", [])
        self.extras = [str(k) + ": " + str(v)
                       for k, v in (attributes_dict.get('extras', {})).items()]
        self.processing = Processing(attributes_dict.get('processing', None))

        self.convert_comments_in_obspy()
        self.obspy_station = self.to_obspy()

    def __repr__(self):
        s = f'\nStation(Label={self.label}, Site={self.site}, '
        s += f'Start Date={self.start_date}, End Date={self.end_date}, '
        s += f'Location Code={self.location_code}, '
        s += f'{len(self.locations)} Locations, '
        if self.processing:
            s += f'processing-steps: {self.processing.processing_list}'
        # if not self.restricted_stations == "unknown":
        #    s += f', {self.restricted_status}'
        s += ')'
        return s

    def to_obspy(self):
        """
        Convert station object to obspy object

        Returns:
            (:class:`obspy.core.inventory.station.Station`):
                  corresponding obspy Station object
        """
        start_date = UTCDateTime(self.start_date) if self.start_date else None
        end_date = UTCDateTime(self.end_date) if self.end_date else None
        site = Site(name=self.site, description=None, town=None, county=None,
                    region=None, country=None)
        comments = [Comment(s) for s in self.comments]

        if self.instrumentation:
            channels_number = len(self.instrumentation.channels)
            chnl_list = [ch.obspy_channel
                         for ch in self.instrumentation.channels]
            equip_list = [self.instrumentation.equipment]
        else:
            channels_number = 0
            chnl_list = []
            equip_list = []

        if self.operator is not None:
            operators = [self.operator.to_obspy()]
        else:
            operators = None
            
        obspy_station = obspy_Station(
            code=self.label,
            latitude=self.location.obspy_latitude,
            longitude=self.location.obspy_longitude,
            elevation=self.location.obspy_elevation,
            channels=chnl_list,
            site=site,
            vault=self.location.vault,
            geology=self.location.geology,
            equipments=equip_list,
            operators=operators,
            creation_date=start_date,
            termination_date=end_date,
            total_number_of_channels=channels_number,
            selected_number_of_channels=channels_number,
            description=None,
            comments=comments,
            start_date=start_date,
            end_date=end_date,
            restricted_status=self.restricted_status,
            alternate_code=None,
            historical_code=None,
            data_availability=None,
            identifiers=None,
            water_level=None,
            source_id=None)
        return obspy_station

    def convert_comments_in_obspy(self):
        """
        Convert info file notes and extras to XML comments
        """

        if self.extras:
            self.comments.append('EXTRA ATTRIBUTES (for documentation only):')
            self.comments = self.comments + self.extras
        if self.processing.processing_list:
            self.comments.append(self.processing.processing_list)



"""
                           "01":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              },
                           "02":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              },
                           "03":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              },
                           {"04":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              }
                           },
                           {"05":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              }
                           },
                           {"06":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              }
                           },
                           {"06":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              }
                           },
                           {"07":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              }
                           },
                           {"08":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              }
                           },
                           {"09":
                             {"position": {"lon": 0., "lat": 0., "elev": 300.},
                              "base":  { 
                                "depth.m": 200.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                                "localisation_method": "Sea surface release point"
                                }
                              }
                           },
                           """
        


"""
Channel, Instrument and Operator classes
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.channel import Channel as obspy_Channel
from obspy.core.inventory.util import (Comment)

# obsinfo modules
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)
from .location import Location
from .orientation import Orientation
from .instrument import Instrument

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Channel(object):
    """
    Corresponds to StationXML/obspy Channel plus channel code

    Attributes:
        das_channel (:class:`ObsMetadata`): represents a channel with defaults
            incorporated
        orientation_code (str): orientation code for this channel (part of
            its ID)
        location_code (str): location code for this channel (part of its ID)
        location (:class:`Location`): represents location corresponding to
            location_code
        start_date (str): inherited from Station
        end_date (str): inherited from Station
        instrument (:class:`Instrument`): a sensor, a datalogger and an
            optional preamplifier
        orientation (:class:`Location`):
        comments (str)
        extras (str)
        obspy_channel (:class:`obspy.core.inventory.channel.Channel`): The
            obspy equivalent  of this object
    """

    def __init__(self, channel_label, attributes_dict, locations,
                 start_date, end_date, equipment, channel_default={},
                 channel_mods_dict={}):
        """
        Constructor

        Args:
            channel_label (str): label to identify channel. Different from the
                channel id code and only used in the information file.
            attributes_dict_or_list (dict/ObsMetadata object or list of
                dicts//ObsMetadata objects): channel attributes
            locations (list of :class:`Location`):
            start_date: (str): using date format
            end_date: (str): using date format
            channel_default (dict or :class:`ObsMetadata`): default attributes
                to complement attributes specified in attributes_dict
            channel_mods_dict (dict or :class:`ObsMetadata`): modification of
                attributes per channel specified in stations
        """
        if channel_label == "default":
            return

        if not attributes_dict:
            msg = 'No channel attributes'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        # Complete das channel fields with default fields. This will create a
        # complete channel
        self.das_channel = ObsMetadata(self.complete_channel(attributes_dict,
                                                             channel_default))
        # Location and orientation code should not be changed, so no channel
        # modifications.  Default is always 00
        self.location_code = self.das_channel.get('location_code', "00")

        # This is just the one-letter string code, Although the attribute
        # is redundant, it si needed before calling class Orientation init
        # at the bottom of this method
        self.orientation_code = self.get_orientation_code(
            attributes_dict.get('orientation_code', None))

        self.location = Location.get_location_from_code(
            locations, self.location_code, "channel", self.channel_id_code)

        self.start_date = start_date
        self.end_date = end_date

        selected_channel_mods = self.get_selected_channel_modifications(
            self.channel_id_code, channel_mods_dict)

        self.instrument = Instrument(self.das_channel, selected_channel_mods)

        # This is the complete dictionary under key orientation code.
        # If present, azimuth and dip can be changed in channel modifications
        orientation_dict = self.das_channel.get('orientation_code', None)
        if isinstance(orientation_dict, str):
            # If just a string, convert to dict to simplify code
            orientation_dict = {orientation_dict: {}}

        orientation_dict = ObsMetadata(orientation_dict)
        # Channel polarity is defined only while creating Instrument object.
        # Don't change assignment order
        self.orientation = Orientation(orientation_dict, selected_channel_mods,
                                       self.instrument.polarity)

        self.comments = attributes_dict.get("comments", [])
        self.extras = [str(k) + ": " + str(v)
                       for k, v in (attributes_dict.get('extras', {})).items()]
        self.convert_notes_and_extras_to_obspy()
        self.obspy_channel = self.to_obspy(equipment)

    @property
    def channel_id_code(self):
        """
        Uniquely identify channel through orientation and location code

        format: {orientation}-{location}

        :returns: channel code
        """
        if not self.orientation_code:
            msg = 'No orientation code in channel'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        # If no location, use 00 as default
        return self.orientation_code + "-" + (self.location_code
                                              if self.location_code else "00")

    def get_orientation_code(self, orientation_dict):
        """
        Get the orientation code from a dict that may include azimuth and dip
        """
        if isinstance(orientation_dict, str):
            keys = orientation_dict
        else:
            keys = list(orientation_dict.keys())
        # There should be only one key
        return keys[0]

    def get_selected_channel_modifications(self, id_code, channel_mods_dict):
        """
        Select a channel_modification by to id_code and channel label.

        An ``id_code`` is composed of an ``orientation code`` plus a "-"
        plus a ``location code``. Selectors for ``channel labels`` in
        modifications can be:

        :"*" or "-": Applies to all ``id_codes``
        :"<NUM>"-"*": Applies to all locations and a given channel
            orientation, <NUM>
        :"*"-"<NUM>": Applies to all orientations and a given channel
            location, <NUM>
        :"<NUM>-"<NUM>": Applies to a give orientation and a given
            location, <NUM>

        Returns:
            selected channel modification
        """
        # Get general default
        default_channel_mod = channel_mods_dict.get("*", {})
        if not default_channel_mod:
            default_channel_mod = channel_mods_dict.get("*-*", {})
        # Get defaults by location and orientation
        default_channel_loc = channel_mods_dict.get(
            self.orientation_code + "-*", {})
        default_channel_orient = channel_mods_dict.get(
            "*-" + self.location_code, {})

        # Get modifications for this particular channel
        chmod = channel_mods_dict.get(id_code, {})
        if not chmod:
            # If id code not found, try with just the orientation part
            chmod = channel_mods_dict.get(id_code[0:1], {})

        # Gather all modifications in a single channel_mods
        # Priority order: particular mods > orientation-specific
        #                 > location-specific > general default
        for k, v in default_channel_loc.items():
            if k not in chmod:
                chmod[k] = v
        for k, v in default_channel_orient.items():
            if k not in chmod:
                chmod[k] = v
        for k, v in default_channel_mod.items():
            if k not in chmod:
                chmod[k] = v
        return chmod

    def complete_channel(self, das_channel, channel_default):
        """
        Complement fields defined for each das channel

        Takes all the fields defined for each das channel and complementa them
        with the channel_default fields. If das_channel key exists, leave the
        corresponding value. If not, add channel_default key/value

        :param das_channel: channel to be completed with default
        :type das_channel: dict or object of :class:`ObsMetadata`
        :param channel_default: default that will complement attributes not
            present in ``das_channel``
        :type channel_default: dict or object of :class:`ObsMetadata`
        """
        # If there are no modifications, use default
        if not das_channel:
            return channel_default

        # First fill out allt he channel attributes
        # Only get channel template if das_channel attribute does not exist
        for k, v in channel_default.items():
            if k not in das_channel:
                das_channel[k] = v

        return das_channel

    def __repr__(self):
        s = f'\n\n\nChannel({self.channel_id_code}, '
        s += f'orientation code="{self.orientation.orientation_code}"'
        if self.location:
            s += f',location={self.location}'

        if self.start_date:
            s += f', startdate={self.start_date}'

        if self.end_date:
            s += f', enddate={self.end_date}'
        s += ')'
        return s

    def channel_code(self, sample_rate):
        """
        Return channel code for a given sample rate.

        Validates instrument and orientation codes according to FDSN
        specifications (for instruments, just the length). Channel codes
        specified by user are indicative and are refined using actual sample
        rate.

        :param sample_rate: instrumentation sampling rate (sps)i
        :type sample_rate: float
        """
        inst_code = self.instrument.sensor.seed_instrument_code

        if len(inst_code) != 1:
            msg = f'Instrument code "{inst_code}" is not a single letter'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)
        if self.orientation.orientation_code not in ["X", "Y", "Z", "1", "2",
                                                     "3", "H", "F"]:
            msg = f'Orientation code "{self.orientation_code}" is not a '\
                  'valid letter'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)

        return (self._band_code_validation(sample_rate)
                + inst_code + self.orientation.orientation_code)

    def _band_code_validation(self, sample_rate):
        """
        Return the channel band code corresponding to a sample rate

        :param sample_rate: sample rate (sps)
        :type sample_rate: float
        """
        bbc = self.instrument.sensor.seed_band_base_code
        if len(bbc) != 1:
            msg = f'Band base code "{bbc}" is not a single letter'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)

        if bbc in "FCHBMLVUWRPTQ":
            if sample_rate >= 1000:
                return "F"
            elif sample_rate >= 250:
                return "C"
            elif sample_rate >= 80:
                return "H"
            elif sample_rate >= 10:
                return "B"
            elif sample_rate > 1:
                return "M"
            elif sample_rate > 0.3:
                return "L"
            elif sample_rate >= 0.1:
                return "V"
            elif sample_rate >= 0.01:
                return "U"
            elif sample_rate >= 0.001:
                return "W"
            elif sample_rate >= 0.0001:
                return "R"
            elif sample_rate >= 0.00001:
                return "P"
            elif sample_rate >= 0.000001:
                return "T"
            else:
                return "Q"
        elif bbc in "GDES":
            if sample_rate >= 1000:
                return "G"
            elif sample_rate >= 250:
                return "D"
            elif sample_rate >= 80:
                return "E"
            elif sample_rate >= 10:
                return "S"
            else:
                msg = "Short period sensor sample rate < 10 sps"
                warnings.warn(msg)
                logger.warning(msg)
        else:
            msg = f'Unknown band base code: "{bbc}"'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

    @property
    def seed_code(self):
        """
        This is equivalent to channel code
        """
        return self.channel_code()

    def to_obspy(self, equipment):
        """
         Create obspy Channel object

         Returns:
            (~class:`obspy.core.inventory.channel.Channel`)
        """
        # FDSN StationXML spec (and therefore obspy) identifies channels by
        # SEED code, different from  obsinfo (first, identification by
        # arbitrary label, then identification by orientation-location
        code = self.channel_code(self.instrument.sample_rate)

        preamp = self.instrument.preamplifier.obspy_equipment \
            if self.instrument.preamplifier else None
        comments = [Comment(s) for s in self.comments]
        equipments = [equipment]  # TFrom instrumentation equipment

        channel = obspy_Channel(
            code, self.location_code,
            latitude=self.location.obspy_latitude,
            longitude=self.location.obspy_longitude,
            elevation=self.location.obspy_elevation,
            depth=self.location.depth_m,
            azimuth=self.orientation.azimuth,
            dip=self.orientation.dip,
            types=['CONTINUOUS', 'GEOPHYSICAL'],
            external_references=None,
            sample_rate=self.instrument.sample_rate,
            sample_rate_ratio_number_samples=None,
            sample_rate_ratio_number_seconds=None,
            storage_format=None,
            clock_drift_in_seconds_per_sample=
                1 / (1e8 * float(self.instrument.sample_rate)),
            calibration_units=None, calibration_units_description=None,
            sensor=self.instrument.sensor.obspy_equipment,
            pre_amplifier=preamp,
            data_logger=self.instrument.datalogger.obspy_equipment,
            equipments=equipments,
            response=self.instrument.obspy_response,
            # description=self.channel_id_code,
            comments=comments,
            start_date=self.start_date,  # OJO: will probably be deprecated
            end_date=self.end_date,
            restricted_status=None,
            alternate_code=None,
            historical_code=None,
            data_availability=None,
            identifiers=None,
            water_level=None,
            source_id=None)
        return channel

    def convert_notes_and_extras_to_obspy(self):
        """
        Convert notes and extras to comments.

        In StationXML comments are found at the channel level and up.
        """
        if self.extras:
            if isinstance(self.extras, list):
                self.comments.extend([f'Extra attribute: {{{e}}}' for e in self.extras])
            else:
                self.comments.append(f'Extra attribute: {{{self.extras}}}')



"""
obsinfo information file routines, contained in superclass ObsMetadata for
generality
"""
# Standard library modules
import json
from pathlib import Path, PurePath
import sys
import re
import warnings
from urllib.parse import urlparse
from urllib.parse import unquote
import logging

# Non-standard modules
import jsonschema
import yaml

# Local modules
from obsinfo.misc import yamlref
from obsinfo.misc.yamlref import JsonLoader
from obsinfo.misc.remoteGitLab import gitLabFile
from obsinfo.misc.discoveryfiles import Datapath

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")

root_symbol = "#"
VALID_FORMATS = ["JSON", "YAML"]
VALID_TYPES = [
    "campaign",
    "network",
    "station",
    "instrumentation",
    "datalogger",
    "preamplifier",
    "sensor",
    "response",
    "stage",
    "filter",
]


class ObsMetadata(dict):
    def __init__(self, *args, **kwargs):
        """
        Constructor, simply create a dict with positional and keyword arguments

        :param args: dictionary of positional arguments
        :param kwars: dictionary of keyword arguments
        """
        super().__init__(*args, **kwargs)

    def convert_to_obsmetadata(self):
        """
        Make all contained dictionaries objects of :class: `.ObsMetadata`
        """

        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = self.__class__(value)
                self[key].convert_to_obsmetadata()
            elif isinstance(value, list):
                for x in value:
                    if isinstance(x, dict):
                        x = self.__class__(x)
                        x.convert_to_obsmetadata()

    def list_valid_types():
        """
        Returns a list of valid information file types

        :returns: list of valid information file types
        """
        return VALID_TYPES

    def is_valid_type(type):
        """
        Returns true if input is a valid schema type
        """
        return type in VALID_TYPES

    def validate(self, schema_path, filename, remote=False, format=None,
                 type=None, verbose=False, schema_file=None, quiet=False,
                 dp=None):
        """
        Validates a YAML or JSON file against schema

        Args:
            schema_path (str): schema path
            filename (str or path-like): name of file to validate
            remote (bool): whether to search for filename in a remote
                repository
            format (str): format of information file:  "JSON" or "YAML"
            type (str): type of info file: "network", "station",
                "instrumentation", "datalogger", "preamplifier", "sensor",
                "stage", "filter"
            verbose (bool): Print progression of validation steps with
                filenames
            schema_file (str): name of schema file
            quiet (bool): Deprecated. Conserved for compatibility. Cancels
                verbose.
            dp (Datapath): datapath for information files.  If None,
                defaults to values stored in .obsinforc

        If type and/or format are not provided, tries to figure them out from
        the filename, which should be "*{TYPE}.{FORMAT}*
        """
        if quiet:
            verbose = False

        if not type and not schema_file:
            type = ObsMetadata.get_information_file_type(filename)

        msg = f"instance = {filename}"
        logger.info(msg)
        if verbose:
            print(msg)

        if dp is None:
            dp = Datapath()  # Datapath for information files, not schemas

        instance = (ObsMetadata.read_json_yaml_ref(filename, dp, format)
                    if not remote else
                    ObsMetadata.read_json_yaml_ref_datapath(filename, dp,
                                                            format))

        base_file = (type if not schema_file else schema_file)
        if not ".schema.json" in base_file:
            base_file += '.schema.json'
        schema_file = PurePath(schema_path).joinpath(base_file)
        base_uri = unquote(PurePath(schema_file).as_uri())

        schema_datapath = Datapath(schema_path)
        try:
            with open(schema_file, "r") as f:
                try:
                    schema = yamlref.loads(
                        f.read(), base_uri=base_uri, jsonschema=True,
                        datapath=schema_datapath, recursive=True)
                except json.decoder.JSONDecodeError as e:
                    msg = ("JSONDecodeError: Error loading JSON schema "
                           f"file: {schema_file}")
                    logger.error(msg)
                    print(msg)
                    logger.error(str(e))
                    print(str(e))
                    return False
                except Exception:
                    msg=f"Error loading JSON schema file: {schema_file}"
                    print(msg)
                    logger.error(msg)
                    print(sys.exc_info()[1])
                    logger.error(sys.exc_info()[1])
                    return False
        except FileNotFoundError:
            msg = f'File not found: {schema_file}'
            logger.error(msg)
            raise FileNotFoundError(msg)
        except (IOError, OSError):
            msg = f'Input/Output error with file: {schema_file}'
            print(msg)
            logger.error(msg)
            raise

        # Lazily report all errors in the instance
        # ASSUMES FIXED SCHEMA (I couldn't get it to work otherwise)
        try:
            msg1 = f"schema =   {schema_file.name}"
            msg2 = "\tTesting schema ..."
            logger.info(msg1)
            logger.info(msg2)
            if verbose:
                print(msg1)
                print(msg2, end="")

            # Check schema first
            jsonschema.Draft7Validator.check_schema(schema)

            msg1 = "OK"
            msg2 = "\tTesting instance ..."
            logger.info(msg1)
            logger.info(msg2)
            if verbose:
                print(msg1)
                print(msg2, end="")

            v = jsonschema.Draft7Validator(schema)

            if not v.is_valid(instance):
                if not quiet:
                    print("")
                errors = sorted(v.iter_errors(instance), key=lambda e: e.path)
                for error in errors:
                    if not quiet:
                        print("\t\t", end="")
                    for elem in error.path:
                        msg = f"['{elem}']"
                        if not quiet:
                            print(msg, end="")
                            logger.error(msg)
                    msg = f": {error.message}"
                    if not quiet:
                        print(msg)
                        logger.error(msg)
                if not quiet:
                    print("\tFAILED")
                    logger.info("\tFAILED")
                return False
            else:
                if verbose:
                    print("OK")
                logger.info("OK")
                return True
        except jsonschema.ValidationError as e:
            logger.error(e.message)
            if not quiet:
                print("\t" + e.message)
            return False

    def get_information_file_format(filename):
        """
        Determines if the information file is in JSON or YAML format.

         Assumes that the filename is "*.{FORMAT}*

        Args:
            filename (str): filename to determine the type of
        Returns:
            format
        Raises:
            (ValueError): on unknown format
        """
        suffix = PurePath(filename).suffix
        format = suffix[1:].upper()
        if format in VALID_FORMATS:
            return format
        msg = f"Unknown format: {format}"
        print(msg)
        logger.error(msg)
        raise ValueError(msg)

    def get_information_file_type(filename):
        """
        Determines the type of a file.

        Assumes that the filename is "*.{TYPE}.{SOMETHING}*

        Args:
            filename (str): filename to determine the type of
        Returns:
            (str): file type
        Raises:
            ValueError
        """
        stem = PurePath(filename).stem
        suffix = PurePath(stem).suffix
        type = suffix[1:]
        if type in VALID_TYPES:
            return type
        msg = f"File '{filename}' is of unknown type: {type}"
        print(msg)
        logger.error(msg)
        raise ValueError(msg)

    def read_json_yaml(filename, format=None):
        """
        Reads a JSON or YAML file. Does NOT use jsonReference  DEPRECATED.

        DEPRECATED. Not being used by any obsinfo method or function. Kept
        for compatibility

        Args:
            filename (str): filename
            format (str): format type, either "YAML" or "JSON"
        Returns:
            (dict):  JSON or YAML parsed information files
        Raises:
            (JSONDecodeError): problem with JSON read
            (FileNotFoundError): file not found
            (IOError): File input/output erre
        """
        if not format:  # validate format is legal. Otherwise program will exit
            format = ObsMetadata.get_information_file_format(filename)

        with open(filename, "r") as f:
            if format == "YAML":
                try:
                    element = yaml.safe_load(f)
                except Exception:
                    msg = f"Error loading YAML file: {filename}"
                    print(msg)
                    logger.error(msg)
                    raise ValueError
            else:
                try:
                    element = json.load(f)
                except json.JSONDecodeError as e:
                    msg = ("JSONDecodeError: Error loading JSON file: "
                           f"{filename}: {str(e)}")
                    print(msg)
                    logger.error(msg)
                except Exception:
                    msg = f"Error loading JSON file: {filename}"
                    print(msg)
                    logger.error(msg)
                    raise
        return element

    def read_json_yaml_ref_datapath(filename, datapath, format=None):
        """
        Reads a JSON or YAML file using jsonReference using OBSINFO_DATAPATH

        Args:
            filename (str): filename
            datapath (:class:`.Datapath`): list of directories to search
                for info files
            format(str): format type, either "YAML" or "JSON"
        Returns:
            (dict):  JSON or YAML parsed information files
        Raises:
            (JSONDecodeError): problem with JSON read
            (FileNotFoundError): file not found
            (IOError): File input/output erre
        """

        if not format:
            format = ObsMetadata.get_information_file_format(filename)

        bu = unquote(filename)

        if gitLabFile.isRemote(bu):
            base_uri = unquote(urlparse(bu).path)
            loader = JsonLoader()
            jsonstr = loader.get_json_or_yaml(bu, base_uri=base_uri,
                                              datapath=datapath)
            return yamlref.loads(jsonstr, base_uri=base_uri, datapath=datapath)
        else:
            base_uri = Path(bu).as_uri()
            try:
                with open(unquote(filename), "r") as f:
                    return yamlref.load(f, base_uri=base_uri,
                                        datapath=datapath)
            except FileNotFoundError:
                msg = f'File not found: {filename}'
                logger.error(msg)
                raise FileNotFoundError(msg)
            except (IOError, OSError):
                msg = f'Input/Output error with file: {filename}'
                print(msg)
                logger.error(msg)
                raise

    def read_json_yaml_ref(filename, datapath, format=None):
        """
        Reads a JSON or YAML file using jsonReference

        Like read_json_yaml_ref, but does not look for files in
        OBSINFO_DATAPATH
        $ref within the data files without absolute or relative path will be
        still looked for in OBSINFO_DATAPATH

        Args:
            filename (str): filename
            datapath (:class:`.Datapath`): object to store list of
            directories to search info files. Used as a dummy.
            format (str): format type, either "YAML" or "JSON"
        Returns:
            (dict): JSON or YAML parsed information files
        Raises:
            (JSONDecodeError): problem with JSON read
            (FileNotFoundError): file not found
            (IOError): File input/output erre
        """
        if not format:
            format = ObsMetadata.get_information_file_format(filename)

        bu = unquote(filename)
        base_uri = Path(bu).as_uri()

        try:
            with open(filename, "r") as f:
                return yamlref.load(f, base_uri=base_uri, datapath=datapath)
        except FileNotFoundError:
            msg = f'File not found: {filename}'
            print(msg)
            logger.error(msg)
            raise FileNotFoundError(msg)
        except (IOError, OSError):
            msg = f'Input/Output error with file: {filename}'
            print(msg)
            logger.error(msg)
            raise

    @staticmethod
    def read_info_file(filename, datapath, remote=False, validate=True,
                       format=None, verbose=False):
        """
        Reads an information file

        Args:
            filename (str): filename
            datapath (:class:`.Datapath`): stores list of directories to
                search info files
            validate (bool): validate before reading
            remote (bool): whether to use absolute/relative path locally
                or OBSINFO_DATAPATH
            format (str): format type ("YAML" or "JSON")

        Returns:
            (:class:`ObsMetadata`): JSON or YAML parsed info files
        """
        if validate:
            if verbose:
                print(f'Validating network file: {filename}')
            logger.info(f'Validating network file: {filename}')
            ObsMetadata().validate(
                Path(__file__).parent.parent.joinpath('data', 'schemas'),
                str(filename), remote=remote, verbose=verbose)
        dict = (ObsMetadata.read_json_yaml_ref(filename, datapath, format)
                if not remote else
                ObsMetadata.read_json_yaml_ref_datapath(filename, datapath,
                                                        format))
        # Create an ObsMedatada object, which is a dir with some added methods
        return ObsMetadata(dict)

#     def _validate_script(argv=None):
#         """
#         Validate an obsinfo information file
#
#         Validates a file named *.{TYPE}.json* or *.{TYPE}.yaml* against the
#         obsinfo schema.{TYPE}.json file.
#
#         {TYPE} can be campaign, network, instrumentation,
#             instrument_components or response
#         """
#         from argparse import ArgumentParser
#
#         parser = ArgumentParser(prog="obsinfo-validate", description=__doc__)
#         parser.add_argument("info_file", help="Information file")
#         parser.add_argument("-t", "--type", choices=VALID_TYPES,default=None,
#                             help="Forces information file type (overrides "
#                                  "interpreting from filename)")
#         parser.add_argument("-f", "--format", choices=VALID_FORMATS,
#                             default=None,
#                             help="Forces information file format (overrides "
#                                  "interpreting from filename)")
#         parser.add_argument("-s", "--schema", default=None,
#                             help="Schema file (overrides interpreting from "
#                                  "filename)")
#         parser.add_argument("-v", "--verbose", action="store_true",
#                             help="increase output verbosity")
#         args = parser.parse_args()
#         validate(args.info_file, format=args.format, type=args.type,
#                  schema_file=args.schema, verbose=args.verbose)

    def get_configured_element(self, key, channel_modification={},
                               selected_configuration={}, default=None):
        """
        Substitute an ObsMetadata element with selected or modification values

        This is the heart of *obsinfo* flexibility, which permits changes in
        configuration or even in individual attributes via channel
        modifications.

        First converts any possible dictionary into an ObsMetadata object to
        be able to use this method recursively

        Returns the value corresponding to key in the following way:

        a) If key present in channel_modification, return the corresponding
           value
        b) Else if key present in selected_configuration, return the
           corresponding value
        c) Else if key present in self, return the corresponding value
        d) Else if key not present in either, return default

        Thus, channel_modifications override selected_configuration which
        overrides self (usually the attributes **read** from an information
        file, which, if not present, are substituted by a default.

        Args:
            key (str): key to values from the respective dictionaries
            channel_modification (dict or :class:`.ObsMetadata`): modifications
                specified at station level
            selected_configuration (dict or :class:`.ObsMetadata`):
                configuration specified at component level
            default (str): default value if none is found
        """
        # If no original value, use default step (c)
        value = self.get(key, default)

        # Use selected_configuration if present, else use original_value
        # (which may be default) (a) and (b) if here is for validation only
        if isinstance(selected_configuration, (dict, ObsMetadata)):
            value = selected_configuration.get(key, value)
        # Use channel_modification if present, else use original_value (which
        # may be default) (a) and (b)
        # Channel modifications override selected configurations which
        # override self which overrides default, but ONLY at the lowest (leaf)
        # level, so we check whethere the value is a dictionary.
        # If so, it won't be substituted. This does not happen with
        # configurations as they are complete substitutions of things such as
        # response stages or sample rates
        if isinstance(channel_modification, (dict, ObsMetadata)):
            val = channel_modification.get(key, value)
            # Do not assign value if this is a dictionary, so it's not a leaf
            # in the hierarchy
            if not isinstance(val, (dict, ObsMetadata)):
                value = val

        if isinstance(value, dict):  # Convert to ObsMetadata any dictionaries
            value = ObsMetadata(value)

        return value

    def validate_dates(dates):
        """
        Converts dates to a standard format

        Args:
            dates (list): dates as strings
        Returns:
            (list): formatted dates as strings
        """

        if dates is None or not isinstance(dates, list):
            return []

        return [ObsMetadata.validate_date(dt) for dt in dates]

    def validate_date(date):
        """
        Reformats an individual date

        Uses regular expressions to match known dates, either in UTC date
        format, in UTC date and time format or in regular dd-mm-yyyy format.
        If only two digits of year specified, assumes 21st century.
        The separator can be either "/" or "-"

        Args:
            date (str): a date in a given format
        Returns:
            (str): a reformatted date as string or None if no value
        Raises:
            (ValueError) if date is unrecognizable
        """

        if date is None or date == int(0):
            # 0 is sometimes the default, the epoch date, 1/1/1970
            return None

        regexp_date_UTC = re.compile("^[0-9]{4}[\-\/][0-1]{0,1}[0-9][\-\/][0-3]{0,1}[0-9]")
        regexp_date_and_time_UTC = re.compile("^[0-9]{4}[\-\/][0-1]{0,1}[0-9][\-\/][0-3]{0,1}[0-9]T[0-2][0-9]:[0-6][0-9]:{0,1}[0-6]{0,1}[0-9]{0,1}Z{0,1}")
        if (re.match(regexp_date_UTC, date)
            or re.match(regexp_date_and_time_UTC, date)):
                return date
        else:  # Assume it's a regular date
            warnings.warn(f'Date {date} is not UTC format, assuming regular '
                          'dd/mm/yyyy format')
            regexp_date_normal = re.compile("[\-\/]")
            date_elements = re.split(regexp_date_normal, date)
            if len(date_elements) != 3:
                raise ValueError("Unrecognizable date, must either be UTC or "
                                 "dd/mm/yyyy or dd/mm/yy. Dashes can be used "
                                 "as separators")
            if len(date_elements[2]) == 2:
                date_elements[2] = "20" + date_elements[2]
                warnings.warn(f'Date {date} is two digits, assuming 21st century')

            return f'{date_elements[2]}-{date_elements[1]}-{date_elements[0]}'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Functions to print obsinfo objects

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import os
from pathlib import Path, PurePath
import glob
import unittest
import inspect
import difflib
import re
import logging

#Third party
from obspy.core.utcdatetime import UTCDateTime
# from pprint import pprint
# import xml.etree.ElementTree as ET
# from CompareXMLTree import XmlTree
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)
from ..instrumentation import (Instrumentation, InstrumentComponent,
                                     Datalogger, Preamplifier, Sensor,
                                     ResponseStages, Stage, Filter, Location)
from ..network import (Station, Network)
from ..instrumentation.filter import (Filter, PolesZeros, FIR, Coefficients, ResponseList,
                     Analog, Digital, AD_Conversion)


class PrintObs(object):
    """
    Collection of methods to print obsinfo objects at different levels of depth.
    All methods are static.
    
      **Attributes:**
     
        *None*
    
    """
    
    @staticmethod
    def print_component(obj, level="all"):
        """
        Prints comoponent information and continues if level is not "component" or "channel".
        
        If level is not "channel" detailed equipment information is not printed.
        
        :param level: determines to which level the *obsinfo* object will be printed
        
        """
        
        if not obj:
            return
        print(obj)
        if level == "channel":
            return 
        
        print(obj.equipment)
        
    @staticmethod
    def print_instrumentation(obj, level="all"):
        """
        Prints instrumentation information and continues if level is not "instrumentation".
        
        If level is  "response" or "all" response stages information will be printed (if it exists).
        Recall at this point all the response is gathered in ``instrument.response_stages.sgates``
        If level is "all" or "filter" (i.e. not "response") all information will be printed (only filter is left at this point...)
        
        :param level: determines to which level the *obsinfo* object will be printed
        
        """
        print(obj)
        print("Instrument:\n")
        print(obj.equipment)
        for k in obj.channels:
            print(k)
            if level != "instrumentation":
                print(k.instrument)
                PrintObs.print_component(k.instrument.sensor, level)
                PrintObs.print_component(k.instrument.preamplifier, level)
                PrintObs.print_component(k.instrument.datalogger, level)
                
                if level in  ["response", "all", "filter"] and k.instrument.response_stages:
                    for s in k.instrument.response_stages:
                       print(s)
                       if level in ["all", "filter"]:
                           print(s.filter) 
            
                
            
    
    @staticmethod            
    def print_station(obj, level="all"):
        """
        Prints station information and continues if level is not "station".
        
        :param level: determines to which level the *obsinfo* object will be printed
        """
        
        print(obj)
        print(obj.locations)
        for k, v in obj.locations.items():
            print(f'Location {k}: {v}')
        if level != "station":
            PrintObs.print_instrumentation(obj.instrumentation, level)
    
    @staticmethod        
    def print_network(obj, level="all"):
        """
        Prints network information and continues if level is not "network".
        
        :param level: determines to which level the *obsinfo* object will be printed
        """
        
        print(obj)
        print(obj.operator)
        if level != "network":
            for v in obj.stations:
                PrintObs.print_station(v, level)
                   



"""
InstrumentComponent class and subclasses Sensor, Preamplifier, Datalogger.
Equipment class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules

# obsinfo
from .response_stages import ResponseStages
from .equipment import Equipment
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class InstrumentComponent(object):
    """
    InstrumentComponent class. Is superclass of all component classes.
    No obspy/StationXML equivalent, because they only specify the whole
    sensor+amplifier+datalogger system

    Attributes:
        equipment (:class:`Equipment`)
        response_stages (:class:`ResponseStages`)
        obspy_equipment (:class:`obspy_Equipment`)
        config_description (str): description of configuration to be added
            to equipment description
    """

    def __init__(self, equipment, response_stages=[],
                 config_description=''):
        """
        Constructor. Invoked by dynamic_class_constructor.

        Args:
            equipment (:class:`Equipment`): equipment features of component
            response_stages (:class:`ResponseStages`): Response of component
                as a list of stages
            config_description (str): description of configuration to be added
                to equipment description
        """
        self.equipment = equipment
        self.config_description = config_description

        if not response_stages or not response_stages.stages:
            msg = f'No response stages in {type(self)}'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)
        self.response_stages = response_stages
        if self.config_description:
            self.equipment.description += ' [config: {}]'.format(
                self.config_description)
        self.obspy_equipment = self.equipment.to_obspy()

    def retrieve_configuration(comp_type, attributes_dict,
                               configuration_selection=None):
        """
        Completes the component configuration.

        A configuration selector can be defined in the instrumentation class,
        otherwise a config default is used.  They may be used to select a
        configuration in the list of configuration definitions. If a
        configuration matches the selector configuration its attributes will
        override or add to existing attributes.

        Args:
            comp_type (str): Type of instrument component. Used for messaging
                purposes only
            attributes_dict (dict or :class:`ObsMetadata`): attributes of
                component, hopefully including configuration_default
            configuration_selection (str): label of configuration to be
                selected
        Returns:
            selected configuration, if found. Otherwise, empty dictionary
        """

        if not attributes_dict:
            return {}

        configuration_default = attributes_dict.get('configuration_default',
                                                    None)
        # If there is a selected configuration at the instrumentation level)
        # use it, else use default
        # configuration at component level
        # If no default configuration selected_configuration will be None
        selected_configuration_key = configuration_selection \
            if configuration_selection else configuration_default

        all_configurations = attributes_dict.get('configuration_definitions',
                                                 {})
        if not selected_configuration_key:
            if all_configurations == {}:
                return {}
            else:
                msg = f'No configuration key or default found in {comp_type}.'\
                       ' No configuration definition will be applied'
                warnings.warn(msg)
                logger.warning(msg)
                return {}
        selected_configuration = all_configurations.get(
            selected_configuration_key, {})

        if selected_configuration == {}:
            msg = "{}'s requested configuration ('{}') is not in configuration_definitions: {}".format(
                comp_type, selected_configuration_key,
                [k for k in all_configurations.keys()])
            logger.warning(msg)
            raise TypeError(msg)
        elif selected_configuration == None:  # configuration is there but empty
            selected_configuration = {"configuration_description": selected_configuration_key}
        else:
            if "configuration_description" not in selected_configuration:
                selected_configuration["configuration_description"] = selected_configuration_key

        return selected_configuration

    @staticmethod
    def dynamic_class_constructor(component_type, attributes_dict,
                                  channel_modif={}, config_selector=''):
        """
        Creates an appropriate Instrument_component subclass (Sensor,
        Preamplifier, Datalogger) from an attributes_dict

        Args:
            component_type (str): type of component. Used for selection of
                adecuate class.
            attributes_dict (dict or :class:`ObsMetadata`): component
                attributes
            channel_modif (dict or :class:`ObsMetadata`): channel modifications
                inherited from station
            delay_correction (float or None): delay correction in seconds:
                if a float: set last component stage to this, others to 0.
                if None: all component stage corrections are set = delay
            config_selector (str): selector of configuration coming from
                Instrument

        Returns:
            object of the adequate subclass
        """
        if not attributes_dict.get(component_type, None):
            if component_type == 'preamplifier':  # Only preamps are optional
                return None
            else:
                msg = f'No {component_type}'
                warnings.warn(msg)
                logger.error(msg)
                raise TypeError(msg)

        selected_config = InstrumentComponent.retrieve_configuration(
            component_type, attributes_dict[component_type], config_selector)

        if component_type == 'datalogger':
            theclass = Datalogger
        elif component_type == 'sensor':
            theclass = Sensor
        elif component_type == 'preamplifier':
            theclass = Preamplifier
        else:
            msg = f'Unknown InstrumentComponent "{component_type}"'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)
        obj = theclass.dynamic_class_constructor(
            ObsMetadata(attributes_dict[component_type]),
            channel_modif.get(component_type, {}),
            selected_config)
        return obj

    def __repr__(self):
        s = ''
        if self.equipment.description:
            s += f', description="{self.equipment.description}"'
        if self.response_stages:
            s += f'Response stages: {len(self.response_stages.stages)}'
        return s


class Sensor(InstrumentComponent):
    """
    Sensor Instrument Component. No obspy equivalent

    Attributes:
        seed_band_base_code (str): SEED base code ("B" or "S") indicating
            instrument band.
        seed_instrument_code (str): SEED instrument code

    """
    def __init__(self, equipment, seed_band_base_code, seed_instrument_code,
                 response_stages=[], config_description=''):
        """
        Constructor

        Args:
            equipment (:class:`.Equipment`): Equipment information
            seed_band_base_code (str (len 1)): SEED base code ("B" or "S")
                indicating instrument band.  Must be modified by
                obsinfo to correspond to output sample rate. Actual SEED base
                code is determined by FDSN standard <http://docs.fdsn.org/
                projects/source-identifiers/en/v1.0/channel-codes.html>`
            seed_instrument code(str (len 1)): SEED instrument code, determined
                by `FDSN standard <http://docs.fdsn.org/projects/source-
                identifiers/en/v1.0/channel-codes.html>`
            response_stages (:class:`ResponseStages`): channel modifications
                inherited from station
            config_description (str): the configuration description that was
                selected, to complement component description
        """
        self.seed_band_base_code = seed_band_base_code
        self.seed_instrument_code = seed_instrument_code  # dictionary
        super().__init__(equipment, response_stages, config_description)

    @classmethod
    def dynamic_class_constructor(cls, attributes_dict, channel_modif={},
                                  selected_config={}):
        """
        Create Sensor instance from an attributes_dict

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): attributes of
                component
            channel_modif (dict or :class:`ObsMetadata`): channel modifications
                inherited from station
            selected_config (dict or :class:`ObsMetadata`): the configuration
                description that will override or complement default values
        Returns:
            (:class:`Sensor`)
        """

        if not attributes_dict:
            return None
        if not selected_config:
            # Avoids a syntax error in the yaml file: two consecutive labels
            # with no response stages
            selected_config = {}

        seed_dict = ObsMetadata(attributes_dict).get_configured_element(
            'seed_codes', channel_modif, selected_config, {})

        # The next line of code will totally override response states in
        # attribute_dict IF there is a selected_config with response_stages
        response_stages_list = attributes_dict.get_configured_element(
            'response_stages', {}, selected_config, None)

        response_stages = ResponseStages(
            response_stages_list,
            channel_modif.get('response_modifications', {}),
            selected_config.get('response_modifications', {}),
            None)

        obj = cls(Equipment(ObsMetadata(attributes_dict.get('equipment',
                                                            None)),
                            channel_modif.get('equipment', {}),
                            selected_config.get('equipment', {})),
                  ObsMetadata(seed_dict).get_configured_element(
                    'band_base', channel_modif, selected_config, None),
                  ObsMetadata(seed_dict).get_configured_element(
                    'instrument', channel_modif, selected_config, None),
                  response_stages,
                  attributes_dict.get_configured_element(
                    'configuration_description', channel_modif,
                    selected_config, ''))
        return obj

    def __repr__(self):
        s = ''
        s += f'\nSensor( "band code={self.seed_band_base_code}", '\
             f'"instrument code={self.seed_instrument_code}")'
        s += super().__repr__()
        return s


class Datalogger(InstrumentComponent):
    """
    Datalogger Instrument Component.

    Obspy equivalent is Datalogger, but only contains elements of
    :class:`Equipment`
    """
    def __init__(self, equipment, sample_rate, delay_correction=None,
                 response_stages=[], config_description=''):
        """
        Constructor

        Args:
            equipment (:class:`Equipment`): equipment attributes
            sample_rate (float): sample rate of given configuration. Checked
                against actual sample rate
            delay_correction (float or None): the delay correction of the
                component. If a float, it is applied to the last stage and the
                other stage corrections are set to 0.  If None, each stage's
                correction is set equal to its delay
            response_stages (:class:`ResponseStages`): channel modifications
                inherited from station
            config_description (str): the configuration description that was
                selected, to complement component description
        """
        self.sample_rate = sample_rate
        self.delay_correction = delay_correction
        super().__init__(equipment, response_stages, config_description)

    @classmethod
    def dynamic_class_constructor(cls, attributes_dict, channel_modif={},
                                  selected_config={}):
        """
        Create Datalogger instance from an attributes_dict

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): component
                attributes
            channel_modif (dict or :class:`ObsMetadata`): channel modifications
                inherited from station
            selected_config (dict or :class:`ObsMetadata`): the configuration
                description that will override or complement default values
        Returns:
            (:class:`Datalogger`)
        """
        if not attributes_dict:
            return None
        if not selected_config:
            # Avoids a syntax error in the yaml file: two consecutive labels
            # with no response stages
            selected_config = {}

        sample_rate = attributes_dict.get_configured_element(
            'sample_rate', channel_modif, selected_config, None)
        delay_correction = attributes_dict.get_configured_element(
            'delay_correction', channel_modif, selected_config, None)
        config_description = attributes_dict.get_configured_element(
            'configuration_description', channel_modif, selected_config, '')

        # The next line of code will totally override response states in
        # attribute_dict IF there is a selected_config with response_stages
        response_stages_list = attributes_dict.get_configured_element(
            'response_stages', {}, selected_config, None)

        response_stages = ResponseStages(
                            response_stages_list,
                            channel_modif.get('response_modifications', {}),
                            selected_config.get('response_modifications', {}),
                            delay_correction)

        obj = cls(Equipment(ObsMetadata(attributes_dict.get('equipment',
                                                            None)),
                            channel_modif.get('equipment', {}),
                            selected_config.get('equipment', {})),
                  sample_rate,
                  delay_correction,
                  response_stages,
                  config_description)

        return obj

    def __repr__(self):
        s = f'Datalogger(Sample Rate={self.sample_rate})'
        s += super().__repr__()
        return s


class Preamplifier(InstrumentComponent):
    """
    Preamplifier Instrument Component. No obspy equivalent

      Attributes:
        None
    """

    def __init__(self, equipment,  response_stages=[],
                 config_description=''):
        """
        Constructor

        Args:
            equipment (:class:`Equipment`): Equipment information
            response_stages (:class:`ResponseStages`): channel modifications
                inherited from station
            config_description (str): the configuration description that was
                selected, added to equipment description
        """
        super().__init__(equipment, response_stages, config_description)

    @classmethod
    def dynamic_class_constructor(cls, attributes_dict, channel_modif={},
                                  selected_config={}):
        """
        Create Preamplifier instance from an attributes_dict

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): attributes of
                component
            channel_modif (dict or :class:`ObsMetadata`): channel modifications
                inherited from station
            selected_config (dict or :class:`ObsMetadata`): the configuration
                description that will override or complement default values
        Returns:
            (:class:`Preamplifier`)
        """

        if not attributes_dict:
            return None
        if not selected_config:
            # Avoids a syntax error in the yaml file: two consecutive labels
            # with no response stages
            selected_config = {}

        # The next line of code will totally override response states in
        # attribute_dict IF there is a selected_config with response_stages
        response_stages_list = attributes_dict.get_configured_element(
            'response_stages', {}, selected_config, None)
        config_description = attributes_dict.get_configured_element(
            'configuration_description', channel_modif, selected_config, '')

        response_stages = ResponseStages(
            response_stages_list,
            channel_modif.get('response_modifications', {}),
            selected_config.get('response_modifications', {}),
            None)

        obj = cls(Equipment(ObsMetadata(attributes_dict.get('equipment',
                                                            None)),
                            channel_modif.get('equipment', {}),
                            selected_config.get('equipment', {})),
                  response_stages,
                  config_description)

        return obj

    def __repr__(self):
        s = 'Preamplifier()'
        s += super().__repr__()
        return s
